# coding=utf-8
import pytest

from ladybug_geometry.geometry3d.arc import Arc3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.polyline import Polyline3D

from ladybug_geometry.geometry2d.arc import Arc2D

import math


def test_arc3_init():
    """Test the initalization of Arc3D objects and basic properties."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)
    str(arc)  # test the string representation of the arc

    assert isinstance(arc.arc2d, Arc2D)
    assert arc.c == pt
    assert arc.radius == arc.arc2d.r == 1
    assert arc.a1 == arc.arc2d.a1 == 0
    assert arc.a2 == arc.arc2d.a2 == math.pi
    assert arc.p1 == Point3D(3, 0, 2)
    assert arc.p2.x == pytest.approx(1, rel=1e-3)
    assert arc.p2.y == pytest.approx(0, rel=1e-3)
    assert arc.p2.z == pytest.approx(2, rel=1e-3)
    assert arc.midpoint == Point3D(2, 1, 2)
    assert arc.length == pytest.approx(math.pi, rel=1e-3)
    assert arc.angle == pytest.approx(math.pi, rel=1e-3)
    assert arc.is_circle is False
    assert arc.is_inverted is False

    arc2 = arc.duplicate()
    assert arc2.c == pt
    assert arc2.radius == 1
    assert arc2.a1 == 0
    assert arc2.a2 == math.pi


def test_arc3_init_radius():
    """Test the initalization of Arc3D objects with a non-unit radius."""
    pt = Point3D(2, 2, 2)
    n = Vector3D(0, 1, 0)
    arc = Arc3D(Plane(n=n, o=pt), 2, math.pi, 0)

    assert arc.c == pt
    assert arc.radius == arc.arc2d.r == 2
    assert arc.a1 == arc.arc2d.a1 == math.pi
    assert arc.a2 == arc.arc2d.a2 == 0
    assert arc.p1.x == pytest.approx(0, rel=1e-3)
    assert arc.p1.z == pytest.approx(2, rel=1e-3)
    assert arc.p2.x == pytest.approx(4, rel=1e-3)
    assert arc.p2.z == pytest.approx(2, rel=1e-3)
    assert arc.midpoint.x == pytest.approx(2, rel=1e-3)
    assert arc.midpoint.z == pytest.approx(4, rel=1e-3)
    assert arc.length == pytest.approx(2 * math.pi, rel=1e-3)
    assert arc.angle == pytest.approx(math.pi, rel=1e-3)
    assert arc.is_circle is False
    assert arc.is_inverted is True


def test_arc3_init_reveresed():
    """Test the initalization of Arc3D objects with reversed direction."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 1.5 * math.pi, 0.5 * math.pi)

    assert arc.c == pt
    assert arc.radius == 1
    assert arc.a1 == 1.5 * math.pi
    assert arc.a2 == 0.5 * math.pi
    assert arc.p1.x == pytest.approx(2, rel=1e-3)
    assert arc.p1.y == pytest.approx(-1, rel=1e-3)
    assert arc.p1.z == pytest.approx(2, rel=1e-3)
    assert arc.p2.x == pytest.approx(2, rel=1e-3)
    assert arc.p2.y == pytest.approx(1, rel=1e-3)
    assert arc.p2.z == pytest.approx(2, rel=1e-3)
    assert arc.midpoint.x == pytest.approx(3, rel=1e-3)
    assert arc.midpoint.y == pytest.approx(0, rel=1e-3)
    assert arc.length == pytest.approx(math.pi, rel=1e-3)
    assert arc.angle == pytest.approx(math.pi, rel=1e-3)
    assert arc.is_circle is False
    assert arc.is_inverted is True


def test_arc3_init_circle():
    """Test the initalization of Arc3D objects as a circle."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 3)

    assert arc.c == pt
    assert arc.radius == 3
    assert arc.a1 == 0
    assert arc.a2 == 2 * math.pi
    assert arc.is_circle
    assert arc.length == math.pi * 6
    assert arc.area == math.pi * 3 ** 2


def test_arc3_init_from_start_mid_end():
    """Test the initalization of Arc3D objects from_start_mid_end."""
    p1 = Point3D(3, 0, 2)
    m = Point3D(2, 1, 2)
    p2 = Point3D(1, 0, 2)
    arc = Arc3D.from_start_mid_end(p1, m, p2)

    assert arc.c == Point3D(2, 0, 2)
    assert arc.radius == 1
    assert arc.a1 == 0
    assert arc.a2 == math.pi
    assert arc.p1 == Point3D(3, 0, 2)
    assert arc.p2.x == pytest.approx(1, rel=1e-3)
    assert arc.p2.y == pytest.approx(0, rel=1e-3)
    assert arc.p2.z == pytest.approx(2, rel=1e-3)
    assert arc.midpoint == Point3D(2, 1, 2)
    assert arc.length == pytest.approx(math.pi, rel=1e-3)
    assert arc.angle == pytest.approx(math.pi, rel=1e-3)
    assert arc.is_circle is False
    assert arc.is_inverted is False

    p1 = Point3D(0, 0)
    m = Point3D(0, 2)
    p2 = Point3D(-1, 1)
    arc = Arc3D.from_start_mid_end(p1, m, p2)

    assert arc.c == Point3D(0, 1)
    assert arc.radius == 1
    assert arc.length == pytest.approx(1.5 * math.pi, rel=1e-3)
    assert arc.is_inverted is True

    p1 = Point3D(0, 0)
    m = Point3D(1, 0)
    p2 = Point3D(2, 0)
    with pytest.raises(Exception):
        arc = Arc3D.from_start_mid_end(p1, m, p2)

def test_arc3_init_from_start_mid_end_flipped():
    """Test the init of Arc3D objects from_start_mid_end using a flipped plane."""
    p1 = Point3D(-91.47, 40.40, 0.76)
    m = Point3D(0.06, 23.23, 97.27)
    p2 = Point3D(91.49, 40.35, 0.99)
    arc = Arc3D.from_start_mid_end(p1, m, p2)

    assert arc.midpoint.z > 0


def test_arc3_to_from_dict():
    """Test the initalization of Arc3D objects and basic properties."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)
    arc_dict = arc.to_dict()
    new_arc = Arc3D.from_dict(arc_dict)
    assert isinstance(new_arc, Arc3D)
    assert new_arc.to_dict() == arc_dict

    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 3)
    arc_dict = arc.to_dict()
    new_arc = Arc3D.from_dict(arc_dict)
    assert isinstance(new_arc, Arc3D)
    assert new_arc.is_circle
    assert new_arc.to_dict() == arc_dict


def test_equality():
    """Test the equality of Arc3D objects."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)
    arc_dup = arc.duplicate()
    arc_alt = Arc3D(Plane(o=Point3D(2, 0.1, 2)), 1, 0.1, math.pi)

    assert arc is arc
    assert arc is not arc_dup
    assert arc == arc_dup
    assert hash(arc) == hash(arc_dup)
    assert arc != arc_alt
    assert hash(arc) != hash(arc_alt)


def test_move():
    """Test the Arc3D move method."""
    pt = Point3D(2, 0, 0)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)

    vec_1 = Vector3D(2, 2, 2)
    new_arc = arc.move(vec_1)
    assert new_arc.c == Point3D(4, 2, 2)
    assert new_arc.radius == arc.radius
    assert new_arc.p1 == Point3D(5, 2, 2)
    assert new_arc.p2 == Point3D(3, 2, 2)
    assert new_arc.a1 == arc.a1
    assert new_arc.a2 == arc.a2


def test_scale():
    """Test the Arc3D scale method."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)

    origin_1 = Point3D(2, 0, 2)
    new_arc = arc.scale(2, origin_1)
    assert new_arc.c == Point3D(2, 0, 2)
    assert new_arc.radius == 2
    assert new_arc.length == 2 * arc.length
    assert new_arc.a1 == arc.a1
    assert new_arc.a2 == arc.a2


def test_scale_world_origin():
    """Test the Arc3D scale method with None origin."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)

    new_arc = arc.scale(2)
    assert new_arc.c == Point3D(4, 0, 4)
    assert new_arc.radius == 2
    assert new_arc.length == 2 * arc.length
    assert new_arc.a1 == arc.a1
    assert new_arc.a2 == arc.a2


def test_rotate():
    """Test the Arc3D rotate method."""
    pt = Point3D(2, 2, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)
    axis_1 = Vector3D(0, 1, 0)
    origin_1 = Point3D(0, 2, 2)

    test_1 = arc.rotate(axis_1, math.pi, origin_1)
    assert test_1.c.x == pytest.approx(-2, rel=1e-3)
    assert test_1.c.z == pytest.approx(2, rel=1e-3)
    assert test_1.radius == arc.radius

    test_2 = arc.rotate(axis_1, math.pi/2, origin_1)
    assert test_2.c.x == pytest.approx(0, rel=1e-3)
    assert test_2.c.z == pytest.approx(0, rel=1e-3)
    assert test_2.radius == arc.radius


def test_rotate_xy():
    """Test the Arc3D rotate_xy method."""
    pt = Point3D(2, 2, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)
    origin_1 = Point3D(0, 2, 2)

    test_1 = arc.rotate_xy(math.pi, origin_1)
    assert test_1.c.x == pytest.approx(-2, rel=1e-3)
    assert test_1.c.y == pytest.approx(2, rel=1e-3)
    assert test_1.radius == arc.radius

    test_2 = arc.rotate_xy(math.pi/2, origin_1)
    assert test_2.c.x == pytest.approx(0, rel=1e-3)
    assert test_2.c.y == pytest.approx(4, rel=1e-3)
    assert test_2.radius == arc.radius


def test_reflect():
    """Test the Arc3D reflect method."""
    pt = Point3D(2, 2, 0)
    arc = Arc3D(Plane(o=pt), 2, 0, math.pi)

    origin_1 = Point3D(0, 1)
    origin_2 = Point3D(1, 1)
    normal_1 = Vector3D(0, 1)
    normal_2 = Vector3D(-1, 1).normalize()
    normal_3 = Vector3D(1, 0)

    assert arc.reflect(normal_1, origin_1).c.x == pytest.approx(2, rel=1e-3)
    assert arc.reflect(normal_1, origin_1).c.y == pytest.approx(0, rel=1e-3)
    assert arc.reflect(normal_1, origin_1).midpoint.x == pytest.approx(2, rel=1e-3)
    assert arc.reflect(normal_1, origin_1).midpoint.y == pytest.approx(-2, rel=1e-3)
    assert arc.reflect(normal_1, origin_2).c == Point3D(2, 0)
    assert arc.reflect(normal_1, origin_2).midpoint.x == pytest.approx(2, rel=1e-3)
    assert arc.reflect(normal_1, origin_2).midpoint.y == pytest.approx(-2, rel=1e-3)

    test_1 = arc.reflect(normal_2, origin_2)
    assert test_1.c.x == pytest.approx(2, rel=1e-3)
    assert test_1.c.y == pytest.approx(2, rel=1e-3)
    assert test_1.midpoint.x == pytest.approx(4, rel=1e-3)
    assert test_1.midpoint.y == pytest.approx(2, rel=1e-3)

    test_2 = arc.reflect(normal_3, origin_1)
    assert test_2.c.x == pytest.approx(-2, rel=1e-3)
    assert test_2.c.y == pytest.approx(2, rel=1e-3)
    assert test_2.midpoint.x == pytest.approx(-2, rel=1e-3)
    assert test_2.midpoint.y == pytest.approx(4, rel=1e-3)


def test_subdivide():
    """Test the Arc3D subdivide methods."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 1.5 * math.pi, 0.5 * math.pi)

    divisions = arc.subdivide(0.5)
    assert len(divisions) == 8
    divisions = arc.subdivide([1, 0.5, 0.25])
    assert len(divisions) == 10
    divisions = arc.subdivide_evenly(4)
    assert len(divisions) == 5


def test_to_polyline():
    """Test the Arc3D to_polyline methods."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 1.5 * math.pi, 0.5 * math.pi)

    pline = arc.to_polyline(4, True)
    assert isinstance(pline, Polyline3D)
    assert pline.vertices[0] == arc.p1
    assert pline.vertices[-1] == arc.p2
    assert len(pline.vertices) == 5
    assert pline.interpolated


def test_point_at_angle():
    """Test the Arc3D point_at_angle method."""
    pt = Point3D(2, 0)
    arc = Arc3D(Plane(o=pt), 1, 1.5 * math.pi, 0.5 * math.pi)

    assert arc.point_at_angle(0.5 * math.pi).x == pytest.approx(3, rel=1e-2)
    assert arc.point_at_angle(0.5 * math.pi).y == pytest.approx(0, rel=1e-2)


def test_arc3_closest_point():
    """Test the Arc3D closest_point method."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)

    assert arc.closest_point(Point3D(2, 2)) == Point3D(2, 1, 2)
    assert arc.closest_point(Point3D(3, 1)).x == pytest.approx(2.71, rel=1e-2)
    assert arc.closest_point(Point3D(3, 1)).y == pytest.approx(0.71, rel=1e-2)
    assert arc.closest_point(Point3D(3, -1)) == Point3D(3, 0, 2)
    assert arc.closest_point(Point3D(1, -1)).x == pytest.approx(1, rel=1e-2)
    assert arc.closest_point(Point3D(1, -1)).y == pytest.approx(0, rel=1e-2)

    assert arc.distance_to_point(Point3D(2, 2)) == math.sqrt(5)
    assert arc.distance_to_point(Point3D(3, 1)) == pytest.approx(2.0424428, rel=1e-2)
    assert arc.distance_to_point(Point3D(3, -1, 2)) == pytest.approx(1, rel=1e-2)
    assert arc.distance_to_point(Point3D(1, -1, 2)) == pytest.approx(1, rel=1e-2)


def test_arc3_intersect_with_plane():
    """Test the Arc3D intersect_with_plane method."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)
    circle = Arc3D(Plane(o=pt), 1)

    plane_1 = Plane(Vector3D(0, 1, 0), Point3D(0, 0, 0))
    int1 = circle.intersect_plane(plane_1)

    assert len(int1) == 2
    assert int1[0].x == pytest.approx(3, rel=1e-2)
    assert int1[0].y == pytest.approx(0, rel=1e-2)
    assert int1[1].x == pytest.approx(1, rel=1e-2)
    assert int1[1].y == pytest.approx(0, rel=1e-2)

    plane_2 = Plane(Vector3D(0, 1, 0), Point3D(0, 0.5, 0))
    int2 = arc.intersect_plane(plane_2)
    assert len(int2) == 2


def test_arc3_split_with_plane():
    """Test the Arc3D split_with_plane method."""
    pt = Point3D(2, 0, 2)
    arc = Arc3D(Plane(o=pt), 1, 0, math.pi)
    circle = Arc3D(Plane(o=pt), 1)

    plane_1 = Plane(Vector3D(0, 1, 0), Point3D(0, 0, 0))
    int1 = circle.split_with_plane(plane_1)
    assert len(int1) == 2
    assert int1[0].p1 == Point3D(3, 0, 2)
    assert int1[0].p2.x == pytest.approx(1, rel=1e-2)
    assert int1[0].p2.y == pytest.approx(0, rel=1e-2)
    assert int1[1].p2 == Point3D(3, 0, 2)
    assert int1[1].p1.x == pytest.approx(1, rel=1e-2)
    assert int1[1].p1.y == pytest.approx(0, rel=1e-2)

    plane_2 = Plane(Vector3D(0, 1, 0), Point3D(0, 0.5, 0))
    int2 = arc.split_with_plane(plane_2)
    assert len(int2) == 3



# coding=utf-8
"""2D Arc"""
from __future__ import division

from .pointvector import Point2D, Vector2D
from .polyline import Polyline2D
from ..intersection2d import closest_point2d_on_arc2d, intersect_line2d_arc2d, \
    intersect_line2d_infinite_arc2d

import math


class Arc2D(object):
    """2D arc object.

    Args:
        c: A Point2D representing the center of the arc.
        r: A number representing the radius of the arc.
        a1: A number between 0 and 2 * pi for the start angle of the arc.
        a2: A number between 0 and 2 * pi for the end angle of the arc.

    Properties:
        * c
        * r
        * a1
        * a2
        * p1
        * p2
        * midpoint
        * length
        * angle
        * is_circle
    """
    __slots__ = ('_c', '_r', '_a1', '_a2', '_cos_a1', '_sin_a1', '_cos_a2', '_sin_a2')

    def __init__(self, c, r, a1=0, a2=2*math.pi):
        """Initialize Arc2D."""
        assert isinstance(c, Point2D), "Expected Point2D. Got {}.".format(type(c))
        assert r > 0, 'Arc radius must be greater than 0. Got {}.'.format(r)
        assert 0 <= a1 <= 2 * math.pi, 'Arc start angle must be between 0 and 2*pi. ' \
            'Got {}.'.format(a1)
        assert 0 <= a2 <= 2 * math.pi, 'Arc start angle must be between 0 and 2*pi. ' \
            'Got {}.'.format(a2)
        self._c = c
        self._r = r
        self._a1 = a1
        self._a2 = a2
        self._cos_a1 = math.cos(a1)
        self._sin_a1 = math.sin(a1)
        self._cos_a2 = math.cos(a2)
        self._sin_a2 = math.sin(a2)

    @classmethod
    def from_dict(cls, data):
        """Create a Arc2D from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "Arc2D"
            "c": (10, 0),
            "r": 5,
            "a1": 0,
            "a2": 3.14159
            }
        """
        return cls(Point2D.from_array(data['c']),
                   data['r'], data['a1'], data['a2'])

    @classmethod
    def from_start_mid_end(cls, p1, m, p2, circle=False):
        """Initialize a new arc from start, middle, and end points.

        Note that input points will be assumed to be in counterclockwise order.

        Args:
            p1: The start point of the arc.
            m: Any point along the length of the arc that is not the start or end.
            p2: The end point of the arc.
            circle: Set to True if you would like the output to be a full circle
                defined by the three points instead of an arc with a start and end.
                Default is False.
        """
        for pt in (p1, m, p2):
            assert isinstance(pt, Point2D), "Expected Point2D. Got {}.".format(type(pt))
        e1 = (p1.x ** 2 + p1.y ** 2)
        e2 = (m.x ** 2 + m.y ** 2)
        e3 = (p2.x ** 2 + p2.y ** 2)
        den = 2 * (p1.x * (m.y - p2.y) - p1.y * (m.x - p2.x) + m.x * p2.y - p2.x * m.y)
        try:
            x = -(e1 * (p2.y - m.y) + e2 * (p1.y - p2.y) + e3 * (m.y - p1.y)) / den
            y = -(e1 * (m.x - p2.x) + e2 * (p2.x - p1.x) + e3 * (p1.x - m.x)) / den
        except ZeroDivisionError:
            raise ValueError('Input points {}, {}, {} are colinear and '
                             'cannot define an arc.'.format(p1, m, p2))
        r = math.sqrt((x - p1.x) ** 2 + (y - p1.y) ** 2)
        if circle is True:
            return cls(Point2D(x, y), r)
        else:
            a1 = Vector2D(1, 0).angle_counterclockwise(Vector2D(p1.x - x, p1.y - y))
            a2 = Vector2D(1, 0).angle_counterclockwise(Vector2D(p2.x - x, p2.y - y))
            return cls(Point2D(x, y), r, a1, a2)

    @property
    def c(self):
        """Center point of the circle on which the arc lies."""
        return self._c

    @property
    def r(self):
        """Radius of arc."""
        return self._r

    @property
    def a1(self):
        """Start angle of the arc in radians."""
        return self._a1

    @property
    def a2(self):
        """End angle of the arc in radians."""
        return self._a2

    @property
    def p1(self):
        """Start point."""
        return Point2D(
            self.c.x + self._cos_a1 * self.r, self.c.y + self._sin_a1 * self.r)

    @property
    def p2(self):
        """End point."""
        return Point2D(
            self.c.x + self._cos_a2 * self.r, self.c.y + self._sin_a2 * self.r)

    @property
    def midpoint(self):
        """Midpoint."""
        return self.point_at(0.5)

    @property
    def length(self):
        """The length of the arc."""
        return self.angle * self.r

    @property
    def angle(self):
        """The total angle over the domain of the arc in radians."""
        _diff = self._a2 - self._a1
        return _diff if not self.is_inverted else 2 * math.pi + _diff

    @property
    def area(self):
        """Area of the circle to which the arc belongs."""
        assert self.is_circle, 'Arc must be a closed circle to access "area" property.'
        return math.pi * self.r ** 2

    @property
    def is_circle(self):
        """Boolean for whether the arc is a full circle (True) or not (False)."""
        return self.a1 == 0 and self.a2 == 2 * math.pi

    @property
    def is_inverted(self):
        """Boolean noting whether the end angle a2 is smaller than the start angle a1."""
        return self._a2 < self._a1

    def move(self, moving_vec):
        """Get an arc that has been moved along a vector.

        Args:
            moving_vec: A Vector2D with the direction and distance to move the arc.
        """
        return Arc2D(self.c.move(moving_vec), self.r, self.a1, self.a2)

    def rotate(self, angle, origin):
        """Get a arc that is rotated counterclockwise by a certain angle.

        Args:
            angle: An angle for rotation in radians.
            origin: A Point2D for the origin around which the arc will
                be rotated.
        """
        _a1 = self.a1 + angle
        _a2 = self.a2 + angle
        _a1 = _a1 - 2 * math.pi if _a1 > 2 * math.pi else _a1
        _a2 = _a2 - 2 * math.pi if _a2 > 2 * math.pi else _a2
        return Arc2D(self.c.rotate(angle, origin), self.r, _a1, _a2)

    def reflect(self, normal, origin):
        """Get a arc reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector2D representing the normal vector for the plane across
                which the arc will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point2D representing the origin from which to reflect.
        """
        return Arc2D.from_start_mid_end(self.p2.reflect(normal, origin),
                                        self.midpoint.reflect(normal, origin),
                                        self.p1.reflect(normal, origin))

    def scale(self, factor, origin=None):
        """Scale a arc by a factor from an origin point.

        Args:
            factor: A number representing how much the arc should be scaled.
            origin: A Point2D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0).
        """
        return Arc2D(self.c.scale(factor, origin), self.r * factor, self.a1, self.a2)

    def subdivide(self, distances):
        """Get Point2D values along the arc that subdivide it based on input distances.

        Args:
            distances: A list of distances along the arc at which to subdivide it.
                This can also be a single number that will be repeated until the
                end of the arc.
        """
        if isinstance(distances, (float, int)):
            distances = [distances]
        arc_length = self.length
        dist = distances[0]
        index = 0
        sub_pts = [self.p1]
        while dist < arc_length:
            sub_pts.append(self.point_at_length(dist))
            if index < len(distances) - 1:
                index += 1
            dist += distances[index]
        sub_pts.append(self.p2)
        return sub_pts

    def subdivide_evenly(self, number):
        """Get Point2D values along the arc that divide it into evenly-spaced segments.

        Args:
            number: The number of segments into which the arc will be divided.
        """
        interval = 1 / number
        parameter = interval
        sub_pts = [self.p1]
        while parameter <= 1:
            sub_pts.append(self.point_at(parameter))
            parameter += interval
        return sub_pts 

    def point_at(self, parameter):
        """Get a point at a given fraction along the arc.

        Args:
            parameter: The fraction between the start and end point where the
                desired point lies. For example, 0.5 will yield the midpoint.
        """
        _ang = self._a1 + self.angle * parameter
        _ang = _ang if _ang <= math.pi * 2 else _ang - math.pi * 2
        return Point2D(
            self.c.x + math.cos(_ang) * self.r, self.c.y + math.sin(_ang) * self.r)

    def point_at_angle(self, angle):
        """Get a point at a given angle along the arc.

        Args:
            angle: The angle in radians from the start point along the arc
                to get the Point2D.
        """
        _ang = self._a1 + angle
        _ang = _ang if _ang <= math.pi * 2 else _ang - math.pi * 2
        return Point2D(
            self.c.x + math.cos(_ang) * self.r, self.c.y + math.sin(_ang) * self.r)

    def point_at_length(self, length):
        """Get a point at a given distance along the arc segment.

        Args:
            length: The distance along the arc from the start point where the
                desired point lies.
        """
        return self.point_at(length / self.length)

    def closest_point(self, point):
        """Get the closest Point2D on this object to another Point2D.

        Args:
            point: A Point2D object to which the closest point on this object
                will be computed.

        Returns:
            Point2D for the closest point on this line to the input point.
        """
        return closest_point2d_on_arc2d(point, self)

    def distance_to_point(self, point):
        """Get the minimum distance between this object and the input point.

        Args:
            point: A Point2D object to which the minimum distance will be computed.

        Returns:
            The distance to the input point.
        """
        close_pt = self.closest_point(point)
        return point.distance_to_point(close_pt)

    def intersect_line_ray(self, line_ray):
        """Get the intersection between this Arc2D and another Ray2 or LineSegment2D.

        Args:
            line_ray: Another LineSegment2D or Ray2D or to intersect.

        Returns:
            A list of 2 Point2D objects if a full intersection exists.
            A list with a single Point2D object if the line is tangent or intersects
            only once. None if no intersection exists.
        """
        return intersect_line2d_arc2d(line_ray, self)

    def intersect_line_infinite(self, line_ray):
        """Get the intersection between this Arc2D and an infinitely extending Ray2D.

        Args:
            line_ray: Another LineSegment2D or Ray2D or to intersect.

        Returns:
            A list of 2 Point2D objects if a full intersection exists.
            A list with a single Point2D object if the line is tangent or intersects
            only once. None if no intersection exists.
        """
        return intersect_line2d_infinite_arc2d(line_ray, self)

    def split_line_infinite(self, line_ray):
        """Split this Arc2D in 2-3 using an infinitely extending Ray2D or LineSegment2D.

        Args:
            line_ray: A LineSegment2D or Ray2D that will be extended infinitely for
                intersection.

        Returns:
            A list with 2 or 3 Arc2D objects if the split was successful.
            Will be a list with 1 Arc2D if no intersection exists.
        """
        inters = intersect_line2d_infinite_arc2d(line_ray, self)
        if inters is None:
            return [self]
        elif self.is_circle:
            if len(inters) != 2:
                return [self]
            a1 = self._a_from_pt(inters[0])
            a2 = self._a_from_pt(inters[1])
            return [Arc2D(self.c, self.r, a1, a2), Arc2D(self.c, self.r, a2, a1)]
        elif len(inters) == 1:
            am = self._a_from_pt(inters[0])
            return [Arc2D(self.c, self.r, self.a1, am),
                    Arc2D(self.c, self.r, am, self.a2)]
        elif len(inters) == 2:
            am1 = self._a_from_pt(inters[0])
            am2 = self._a_from_pt(inters[1])
            if self._cc_difference(am1) < self._cc_difference(am2):
                return [Arc2D(self.c, self.r, self.a1, am1),
                        Arc2D(self.c, self.r, am1, am2),
                        Arc2D(self.c, self.r, am2, self.a2)]
            else:
                return [Arc2D(self.c, self.r, self.a1, am2),
                        Arc2D(self.c, self.r, am2, am1),
                        Arc2D(self.c, self.r, am1, self.a2)]

    def to_polyline(self, divisions, interpolated=True):
        """Get this Arc2D as an approximated Polyline2D.
        
        Args:
            divisions: The number of segments into which the arc will be divided.
            interpolated: Boolean to note whether the polyline should be interpolated
                between the input vertices when it is translated to other interfaces.
                This property has no effect on the geometric calculations performed
                by this library and is only present in order to assist with
                display/translation. (Default: True)
        """
        pts = self.subdivide_evenly(divisions)
        return Polyline2D(pts, interpolated)

    def to_dict(self):
        """Get Arc2D as a dictionary."""
        return {'type': 'Arc2D', 'c': self.c.to_array(),
                'r': self.r, 'a1': self.a1, 'a2': self.a2}

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def _pt_in(self, point):
        if self.is_circle:
            return True
        else:
            v = Vector2D(point.x - self.c.x, point.y - self.c.y)
            a = Vector2D(1, 0).angle_counterclockwise(v)
            return (not self.is_inverted and self.a1 < a < self.a2) or \
                (self.is_inverted and self.a1 > a > self.a2)

    def _a_from_pt(self, point):
        """Get the angle along the arc given a point along the arc."""
        v = Vector2D(point.x - self.c.x, point.y - self.c.y)
        return Vector2D(1, 0).angle_counterclockwise(v)

    def _cc_difference(self, angle):
        """Get counterclockwise different between an angle and the start of this arc."""
        _diff = angle - self.a1
        return _diff if not angle < self.a1 else 2 * math.pi + _diff

    def __copy__(self):
        return Arc2D(self.c, self.r, self.a1, self.a2)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.c, self.r, self.a1, self.a2)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Arc2D) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        return 'Arc2D (center <%.2f, %.2f>) (radius <%.2f>) (length <%.2f>)' % \
            (self.c.x, self.c.y, self.r, self.length)



# coding=utf-8
"""2D Ray"""
from __future__ import division

from .pointvector import Vector2D, Point2D
from ._1d import Base1DIn2D


class Ray2D(Base1DIn2D):
    """2D Ray object.

    Args:
        p: A Point2D representing the base of the ray.
        v: A Vector2D representing the direction of the ray.

    Properties:
        * p
        * v
    """
    __slots__ = ()

    def __init__(self, p, v):
        """Initialize Ray2D."""
        Base1DIn2D.__init__(self, p, v)

    @classmethod
    def from_array(cls, ray_array):
        """ Create a Ray2D from a nested array with a point and a vector.

        Args:
            ray_array: Nested tuples ((p.x, p.y), (v.x, v.y)).
        """
        return Ray2D(Point2D(*ray_array[0]), Vector2D(*ray_array[1]))

    def reverse(self):
        """Get a copy of this ray that is reversed."""
        return Ray2D(self.p, self.v.reverse())

    def move(self, moving_vec):
        """Get a ray that has been moved along a vector.

        Args:
            moving_vec: A Vector2D with the direction and distance to move the ray.
        """
        return Ray2D(self.p.move(moving_vec), self.v)

    def rotate(self, angle, origin):
        """Get a ray that is rotated counterclockwise by a certain angle.

        Args:
            angle: An angle for rotation in radians.
            origin: A Point2D for the origin around which the ray will be rotated.
        """
        return Ray2D(self.p.rotate(angle, origin), self.v.rotate(angle))

    def reflect(self, normal, origin):
        """Get a ray reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector2D representing the normal vector for the plane across
                which the ray will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point2D representing the origin from which to reflect.
        """
        return Ray2D(self.p.reflect(normal, origin), self.v.reflect(normal))

    def scale(self, factor, origin=None):
        """Scale a ray by a factor from an origin point.

        Args:
            factor: A number representing how much the ray should be scaled.
            origin: A Point2D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0).
        """
        return Ray2D(self.p.scale(factor, origin), self.v * factor)

    def to_dict(self):
        """Get Ray2D as a dictionary."""
        base = Base1DIn2D.to_dict(self)
        base['type'] = 'Ray2D'
        return base

    def to_array(self):
        """A nested array representing the start point and vector."""
        return (self.p.to_array(), self.v.to_array())

    def _u_in(self, u):
        return u >= 0.0

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self.p), hash(self.v))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Ray2D) and self.__key() == other.__key()

    def __repr__(self):
        return 'Ray2D (point <%.2f, %.2f>) (vector <%.2f, %.2f>)' % \
            (self.p.x, self.p.y, self.v.x, self.v.y)



# coding=utf-8
"""Sphere"""
from __future__ import division

from .pointvector import Point3D
from .plane import Plane
from .arc import Arc3D
from .line import LineSegment3D
from ..intersection3d import intersect_line3d_sphere, intersect_plane_sphere


import math


class Sphere(object):
    """Sphere object.

    Args:
        center: A Point3D representing the center of the arc.
        radius: A number representing the radius of the arc.

    Properties:
        * center
        * radius
        * min
        * max
        * diameter
        * circumference
        * area
        * volume
    """
    __slots__ = ('_center', '_radius')

    def __init__(self, center, radius):
        """Initialize Sphere."""
        assert isinstance(center, Point3D), \
            "Expected Point3D. Got {}.".format(type(center))
        assert radius > 0, 'Sphere radius must be greater than 0. Got {}.'.format(radius)
        self._center = center
        self._radius = radius

    @classmethod
    def from_dict(cls, data):
        """Create a Sphere from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "Sphere"
            "center": (10, 0, 0),
            "radius": 5,
            }
        """
        return cls(Point3D.from_array(data['center']), data['radius'])

    @property
    def center(self):
        """Center of sphere."""
        return self._center

    @property
    def radius(self):
        """Radius of sphere."""
        return self._radius

    @property
    def min(self):
        """A Point3D for the minimum bounding box vertex around this geometry."""
        return Point3D(self.center.x - self.radius, self.center.y - self.radius,
                       self.center.z - self.radius)

    @property
    def max(self):
        """A Point3D for the maximum bounding box vertex around this geometry."""
        return Point3D(self.center.x + self.radius, self.center.y + self.radius,
                       self.center.z + self.radius)

    @property
    def diameter(self):
        """Diameter of sphere"""
        return self.radius * 2

    @property
    def circumference(self):
        """Circumference of sphere"""
        return 2 * math.pi * self.radius

    @property
    def area(self):
        """Surface area of sphere"""
        return 4 * math.pi * self.radius ** 2

    @property
    def volume(self):
        """Volume of sphere"""
        return 4 / 3 * math.pi * self.radius ** 3

    def move(self, moving_vec):
        """Get a sphere that has been moved along a vector.

        Args:
            moving_vec: A Vector3D with the direction and distance to move the sphere.
        """
        return Sphere(self.center.move(moving_vec), self.radius)

    def rotate(self, axis, angle, origin):
        """Rotate this sphere by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis: A Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in radians.
            origin: A Point3D for the origin around which the sphere will be rotated.
        """
        return Sphere(self.center.rotate(axis, angle, origin), self.radius)

    def rotate_xy(self, angle, origin):
        """Get a sphere that is rotated counterclockwise in the world XY plane by an angle.

        Args:
            angle: An angle for rotation in radians.
            origin: A Point3D for the origin around which the sphere will be rotated.
        """
        return Sphere(self.center.rotate_xy(angle, origin), self.radius)

    def reflect(self, normal, origin):
        """Get a sphere reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector3D representing the normal vector for the plane across
                which the arc will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point3D representing the origin from which to reflect.
        """
        return Sphere(self.center.reflect(normal, origin), self.radius)

    def scale(self, factor, origin=None):
        """Scale a sphere by a factor from an origin point.

        Args:
            factor: A number representing how much the sphere should be scaled.
            origin: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        return Sphere(self.center.scale(factor, origin), self.radius * factor)

    def intersect_plane(self, plane):
        """Get the intersection of a plane with this Sphere object

        Args:
            plane: A Plane object.

        Returns:
            Arc3D representing a full circle if it exists.
            None if no full intersection exists.
        """
        ip = intersect_plane_sphere(plane, self)  # ip = [center pt, vector, radius]
        return None if ip is None or isinstance(ip, Point3D) else \
            Arc3D(Plane(ip[1], ip[0]), ip[2])

    def intersect_line_ray(self, line_ray):
        """Get the intersection between this Sphere object and a Ray2D/LineSegment2D.

        Args:
            line_ray: A LineSegment3D or Ray3D that will be extended infinitely
                for intersection.

        Returns:
            A LineSegment3D object if a full intersection exists.
            A Point if a tangent intersection exists.
            None if no full intersection exists.
        """
        il = intersect_line3d_sphere(line_ray, self)
        return None if il is None else \
            il if isinstance(il, Point3D) else \
            LineSegment3D.from_end_points(il[0], il[1])

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def to_dict(self):
        """Get Sphere as a dictionary."""
        return {'type': 'Sphere',
                'center': self.center.to_array(),
                'radius': self.radius}

    def __copy__(self):
        return Sphere(self._center, self._radius)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self._center, self._radius)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Sphere) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        return 'Sphere (center {}) (radius {}))'.format(self.center, self.radius)



# coding=utf-8
import pytest

from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.ray import Ray2D
from ladybug_geometry.geometry2d.arc import Arc2D
from ladybug_geometry.geometry2d.polyline import Polyline2D

import math


def test_arc2_init():
    """Test the initalization of Arc2D objects and basic properties."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 0, math.pi)
    str(arc)  # test the string representation of the arc

    assert arc.c == pt
    assert arc.r == 1
    assert arc.a1 == 0
    assert arc.a2 == math.pi
    assert arc.p1 == Point2D(3, 0)
    assert arc.p2.x == pytest.approx(1, rel=1e-3)
    assert arc.p2.y == pytest.approx(0, rel=1e-3)
    assert arc.midpoint == Point2D(2, 1)
    assert arc.length == pytest.approx(math.pi, rel=1e-3)
    assert arc.angle == pytest.approx(math.pi, rel=1e-3)
    assert arc.is_circle is False
    assert arc.is_inverted is False

    arc2 = arc.duplicate()
    assert arc2.c == pt
    assert arc2.r == 1
    assert arc2.a1 == 0
    assert arc2.a2 == math.pi


def test_arc2_init_radius():
    """Test the initalization of Arc2D objects with a non-unit radius."""
    pt = Point2D(2, 2)
    arc = Arc2D(pt, 2,  math.pi, 0)

    assert arc.c == pt
    assert arc.r == 2
    assert arc.a1 == math.pi
    assert arc.a2 == 0
    assert arc.p1.x == pytest.approx(0, rel=1e-3)
    assert arc.p1.y == pytest.approx(2, rel=1e-3)
    assert arc.p2.x == pytest.approx(4, rel=1e-3)
    assert arc.p2.y == pytest.approx(2, rel=1e-3)
    assert arc.midpoint.x == pytest.approx(2, rel=1e-3)
    assert arc.midpoint.y == pytest.approx(0, rel=1e-3)
    assert arc.length == pytest.approx(2 * math.pi, rel=1e-3)
    assert arc.angle == pytest.approx(math.pi, rel=1e-3)
    assert arc.is_circle is False
    assert arc.is_inverted is True


def test_arc2_init_reveresed():
    """Test the initalization of Arc2D objects with reversed direction."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 1.5 * math.pi, 0.5 * math.pi)

    assert arc.c == pt
    assert arc.r == 1
    assert arc.a1 == 1.5 * math.pi
    assert arc.a2 == 0.5 * math.pi
    assert arc.p1.x == pytest.approx(2, rel=1e-3)
    assert arc.p1.y == pytest.approx(-1, rel=1e-3)
    assert arc.p2.x == pytest.approx(2, rel=1e-3)
    assert arc.p2.y == pytest.approx(1, rel=1e-3)
    assert arc.midpoint.x == pytest.approx(3, rel=1e-3)
    assert arc.midpoint.y == pytest.approx(0, rel=1e-3)
    assert arc.length == pytest.approx(math.pi, rel=1e-3)
    assert arc.angle == pytest.approx(math.pi, rel=1e-3)
    assert arc.is_circle is False
    assert arc.is_inverted is True


def test_arc2_init_circle():
    """Test the initalization of Arc2D objects as a circle."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 3)

    assert arc.c == pt
    assert arc.r == 3
    assert arc.a1 == 0
    assert arc.a2 == 2 * math.pi
    assert arc.is_circle
    assert arc.length == math.pi * 6
    assert arc.area == math.pi * 3 ** 2


def test_arc2_init_from_start_mid_end():
    """Test the initalization of Arc2D objects from_start_mid_end."""
    p1 = Point2D(3, 0)
    m = Point2D(2, 1)
    p2 = Point2D(1, 0)
    arc = Arc2D.from_start_mid_end(p1, m, p2)

    assert arc.c == Point2D(2, 0)
    assert arc.r == 1
    assert arc.a1 == 0
    assert arc.a2 == math.pi
    assert arc.p1 == Point2D(3, 0)
    assert arc.p2.x == pytest.approx(1, rel=1e-3)
    assert arc.p2.y == pytest.approx(0, rel=1e-3)
    assert arc.midpoint == Point2D(2, 1)
    assert arc.length == pytest.approx(math.pi, rel=1e-3)
    assert arc.angle == pytest.approx(math.pi, rel=1e-3)
    assert arc.is_circle is False
    assert arc.is_inverted is False

    p1 = Point2D(0, 0)
    m = Point2D(0, 2)
    p2 = Point2D(-1, 1)
    arc = Arc2D.from_start_mid_end(p1, m, p2)

    assert arc.c == Point2D(0, 1)
    assert arc.r == 1
    assert arc.length == pytest.approx(1.5 * math.pi, rel=1e-3)
    assert arc.is_inverted is True

    p1 = Point2D(0, 0)
    m = Point2D(1, 0)
    p2 = Point2D(2, 0)
    with pytest.raises(Exception):
        arc = Arc2D.from_start_mid_end(p1, m, p2)


def test_arc2_to_from_dict():
    """Test the initalization of Arc2D objects and basic properties."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 0, math.pi)
    arc_dict = arc.to_dict()
    new_arc = Arc2D.from_dict(arc_dict)
    assert isinstance(new_arc, Arc2D)
    assert new_arc.to_dict() == arc_dict

    pt = Point2D(2, 0)
    arc = Arc2D(pt, 3)
    arc_dict = arc.to_dict()
    new_arc = Arc2D.from_dict(arc_dict)
    assert isinstance(new_arc, Arc2D)
    assert new_arc.is_circle
    assert new_arc.to_dict() == arc_dict


def test_equality():
    """Test the equality of Arc2D objects."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 0, math.pi)
    arc_dup = arc.duplicate()
    arc_alt = Arc2D(pt, 1, 0.1, math.pi)

    assert arc is arc
    assert arc is not arc_dup
    assert arc == arc_dup
    assert hash(arc) == hash(arc_dup)
    assert arc != arc_alt
    assert hash(arc) != hash(arc_alt)


def test_move():
    """Test the Arc2D move method."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 0, math.pi)

    vec_1 = Vector2D(2, 2)
    new_arc = arc.move(vec_1)
    assert new_arc.c == Point2D(4, 2)
    assert new_arc.r == arc.r
    assert new_arc.p1 == Point2D(5, 2)
    assert new_arc.p2 == Point2D(3, 2)
    assert new_arc.a1 == arc.a1
    assert new_arc.a2 == arc.a2


def test_scale():
    """Test the Arc2D scale method."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 0, math.pi)

    origin_1 = Point2D(2, 0)
    new_arc = arc.scale(2, origin_1)
    assert new_arc.c == Point2D(2, 0)
    assert new_arc.r == 2
    assert new_arc.length == 2 * arc.length
    assert new_arc.a1 == arc.a1
    assert new_arc.a2 == arc.a2


def test_scale_world_origin():
    """Test the Arc2D scale method with None origin."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 0, math.pi)

    new_arc = arc.scale(2)
    assert new_arc.c == Point2D(4, 0)
    assert new_arc.r == 2
    assert new_arc.length == 2 * arc.length
    assert new_arc.a1 == arc.a1
    assert new_arc.a2 == arc.a2


def test_rotate():
    """Test the Arc2D rotate method."""
    pt = Point2D(2, 2)
    arc = Arc2D(pt, 1, 0, math.pi)
    origin_1 = Point2D(0, 2)

    test_1 = arc.rotate(math.pi, origin_1)
    assert test_1.c.x == pytest.approx(-2, rel=1e-3)
    assert test_1.c.y == pytest.approx(2, rel=1e-3)
    assert test_1.r == arc.r

    test_2 = arc.rotate(math.pi/2, origin_1)
    assert test_2.c.x == pytest.approx(0, rel=1e-3)
    assert test_2.c.y == pytest.approx(4, rel=1e-3)
    assert test_2.r == arc.r


def test_reflect():
    """Test the Arc2D reflect method."""
    pt = Point2D(2, 2)
    arc = Arc2D(pt, 2, 0, math.pi)

    origin_1 = Point2D(0, 1)
    origin_2 = Point2D(1, 1)
    normal_1 = Vector2D(0, 1)
    normal_2 = Vector2D(-1, 1).normalize()

    assert arc.reflect(normal_1, origin_1).c.x == pytest.approx(2, rel=1e-3)
    assert arc.reflect(normal_1, origin_1).c.y == pytest.approx(0, rel=1e-3)
    assert arc.reflect(normal_1, origin_1).midpoint.x == pytest.approx(2, rel=1e-3)
    assert arc.reflect(normal_1, origin_1).midpoint.y == pytest.approx(-2, rel=1e-3)
    assert arc.reflect(normal_1, origin_2).c == Point2D(2, 0)
    assert arc.reflect(normal_1, origin_2).midpoint.x == pytest.approx(2, rel=1e-3)
    assert arc.reflect(normal_1, origin_2).midpoint.y == pytest.approx(-2, rel=1e-3)

    test_1 = arc.reflect(normal_2, origin_2)
    assert test_1.c.x == pytest.approx(2, rel=1e-3)
    assert test_1.c.y == pytest.approx(2, rel=1e-3)
    assert test_1.midpoint.x == pytest.approx(4, rel=1e-3)
    assert test_1.midpoint.y == pytest.approx(2, rel=1e-3)


def test_subdivide():
    """Test the Arc2D subdivide methods."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 1.5 * math.pi, 0.5 * math.pi)

    divisions = arc.subdivide(0.5)
    assert len(divisions) == 8
    divisions = arc.subdivide([1, 0.5, 0.25])
    assert len(divisions) == 10
    divisions = arc.subdivide_evenly(4)
    assert len(divisions) == 5


def test_to_polyline():
    """Test the Arc3D to_polyline methods."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 1.5 * math.pi, 0.5 * math.pi)

    pline = arc.to_polyline(4, True)
    assert isinstance(pline, Polyline2D)
    assert pline.vertices[0] == arc.p1
    assert pline.vertices[-1] == arc.p2
    assert len(pline.vertices) == 5
    assert pline.interpolated


def test_point_at_angle():
    """Test the Arc2D point_at_angle method."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 1.5 * math.pi, 0.5 * math.pi)

    assert arc.point_at_angle(0.5 * math.pi).x == pytest.approx(3, rel=1e-2)
    assert arc.point_at_angle(0.5 * math.pi).y == pytest.approx(0, rel=1e-2)


def test_arc2_closest_point():
    """Test the Arc2D closest_point method."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 0, math.pi)

    assert arc.closest_point(Point2D(2, 2)) == Point2D(2, 1)
    assert arc.closest_point(Point2D(3, 1)).x == pytest.approx(2.71, rel=1e-2)
    assert arc.closest_point(Point2D(3, 1)).y == pytest.approx(0.71, rel=1e-2)
    assert arc.closest_point(Point2D(3, -1)) == Point2D(3, 0)
    assert arc.closest_point(Point2D(1, -1)).x == pytest.approx(1, rel=1e-2)
    assert arc.closest_point(Point2D(1, -1)).y == pytest.approx(0, rel=1e-2)

    assert arc.distance_to_point(Point2D(2, 2)) == 1
    assert arc.distance_to_point(Point2D(3, 1)) == pytest.approx(0.4142, rel=1e-2)
    assert arc.distance_to_point(Point2D(3, -1)) == pytest.approx(1, rel=1e-2)
    assert arc.distance_to_point(Point2D(1, -1)) == pytest.approx(1, rel=1e-2)


def test_arc2_intersect_line_ray():
    """Test the Arc2D intersect_line_ray method."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 0, math.pi)
    circle = Arc2D(pt, 1)
    seg1 = LineSegment2D(pt, Vector2D(2, 2))
    seg2 = LineSegment2D(Point2D(0, -2), Vector2D(6, 6))
    seg3 = LineSegment2D(pt, Vector2D(0.5, 0.5))

    int1 = arc.intersect_line_ray(seg1)
    assert len(int1) == 1
    assert int1[0].x == pytest.approx(2.71, rel=1e-2)
    assert int1[0].y == pytest.approx(0.71, rel=1e-2)
    int2 = circle.intersect_line_ray(seg1)
    assert len(int2) == 1
    assert int2[0].x == pytest.approx(2.71, rel=1e-2)
    assert int2[0].y == pytest.approx(0.71, rel=1e-2)
    int3 = arc.intersect_line_ray(seg2)
    assert len(int3) == 1
    assert int3[0].x == pytest.approx(2.71, rel=1e-2)
    assert int3[0].y == pytest.approx(0.71, rel=1e-2)
    int4 = circle.intersect_line_ray(seg2)
    assert len(int4) == 2
    assert int4[0].x == pytest.approx(2.71, rel=1e-2)
    assert int4[0].y == pytest.approx(0.71, rel=1e-2)
    assert int4[1].x == pytest.approx(1.29, rel=1e-2)
    assert int4[1].y == pytest.approx(-0.71, rel=1e-2)
    assert arc.intersect_line_ray(seg3) is None


def test_arc2_split_line_infinite():
    """Test the Arc2D split_line_ray_infinite method."""
    pt = Point2D(2, 0)
    arc = Arc2D(pt, 1, 0, math.pi)
    circle = Arc2D(pt, 1)

    int1 = circle.split_line_infinite(Ray2D(Point2D(), Vector2D(1, 0)))
    assert len(int1) == 2
    assert int1[0].p1 == Point2D(3, 0)
    assert int1[0].p2.x == pytest.approx(1, rel=1e-2)
    assert int1[0].p2.y == pytest.approx(0, rel=1e-2)
    assert int1[1].p2 == Point2D(3, 0)
    assert int1[1].p1.x == pytest.approx(1, rel=1e-2)
    assert int1[1].p1.y == pytest.approx(0, rel=1e-2)

    int2 = arc.split_line_infinite(Ray2D(Point2D(0, 0.5), Vector2D(1, 0)))
    assert len(int2) == 3



# coding=utf-8
"""2D Vector and 2D Point"""
from __future__ import division

import math
import operator


class Vector2D(object):
    """2D Vector object.

    Args:
        x: Number for the X coordinate.
        y: Number for the Y coordinate.

    Properties:
        * x
        * y
        * magnitude
        * magnitude_squared
        * is_zero
    """
    __slots__ = ('_x', '_y')

    def __init__(self, x=0, y=0):
        """Initialize 2D Vector."""
        self._x = self._cast_to_float(x)
        self._y = self._cast_to_float(y)

    @classmethod
    def from_dict(cls, data):
        """Create a Vector2D/Point2D from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "x": 10,
            "y": 0
            }
        """
        return cls(data['x'], data['y'])

    @classmethod
    def from_array(cls, array):
        """Initialize a Vector2D/Point2D from an array.

        Args:
            array: A tuple or list with two numbers representing the x and y
                values of the point.
        """
        return cls(array[0], array[1])

    def to_array(self):
        """Get Vector2D/Point2D as a tuple of two numbers"""
        return (self.x, self.y)

    @property
    def x(self):
        """Get the X coordinate."""
        return self._x

    @property
    def y(self):
        """Get the Y coordinate."""
        return self._y

    @property
    def magnitude(self):
        """Get the magnitude of the vector."""
        return self.__abs__()

    @property
    def magnitude_squared(self):
        """Get the magnitude squared of the vector."""
        return self.x ** 2 + self.y ** 2

    def is_zero(self, tolerance):
        """Boolean to note whether the vector is within a given zero tolerance.

        Args:
            tolerance: The tolerance below which the vector is considered to
                be a zero vector.
        """
        return abs(self.x) <= tolerance and abs(self.y) <= tolerance

    def normalize(self):
        """Get a copy of the vector that is a unit vector (magnitude=1)."""
        d = self.magnitude
        try:
            return Vector2D(self.x / d, self.y / d)
        except ZeroDivisionError:
            return self.duplicate()

    def reverse(self):
        """Get a copy of this vector that is reversed."""
        return self.__neg__()

    def dot(self, other):
        """Get the dot product of this vector with another."""
        return self.x * other.x + self.y * other.y

    def determinant(self, other):
        """Get the determinant between this vector and another 2D vector."""
        return self.x * other.y - self.y * other.x

    def cross(self):
        """Get the cross product of this vector."""
        return Vector2D(self.y, -self.x)

    def angle(self, other):
        """Get the smallest angle between this vector and another."""
        try:
            return math.acos(self.dot(other) / (self.magnitude * other.magnitude))
        except ValueError:  # python floating tolerance can cause math domain error
            if self.dot(other) < 0:
                return math.acos(-1)
            return math.acos(1)

    def angle_counterclockwise(self, other):
        """Get the counterclockwise angle between this vector and another."""
        inner = self.angle(other)
        det = self.determinant(other)
        if det >= 0:
            return inner  # if the det > 0 then self is immediately clockwise of other
        else:
            return 2 * math.pi - inner  # if the det < 0 then other is clockwise of self

    def angle_clockwise(self, other):
        """Get the clockwise angle between this vector and another."""
        inner = self.angle(other)
        det = self.determinant(other)
        if det <= 0:
            return inner  # if the det > 0 then self is immediately clockwise of other
        else:
            return 2 * math.pi - inner  # if the det < 0 then other is clockwise of self

    def rotate(self, angle):
        """Get a vector that is rotated counterclockwise by a certain angle.

        Args:
            angle: An angle for rotation in radians.
        """
        return Vector2D._rotate(self, angle)

    def reflect(self, normal):
        """Get a vector that is reflected across a plane with the input normal vector.

        Args:
            normal: A Vector2D representing the normal vector for the plane across
                which the vector will be reflected. THIS VECTOR MUST BE NORMALIZED.
        """
        return Vector2D._reflect(self, normal)

    def duplicate(self):
        """Get a copy of this vector."""
        return self.__copy__()

    def to_dict(self):
        """Get Vector2D as a dictionary."""
        return {'type': 'Vector2D',
                'x': self.x,
                'y': self.y}

    def _cast_to_float(self, value):
        """Ensure that an input coordinate value is a float."""
        try:
            number = float(value)
        except (ValueError, TypeError):
            raise TypeError(
                'Coordinates must be numbers. Got {}: {}.'.format(type(value), value))
        return number

    @staticmethod
    def _rotate(vec, angle):
        """Hidden rotation method used by both Point2D and Vector2D."""
        cos_a = math.cos(angle)
        sin_a = math.sin(angle)
        qx = cos_a * vec.x - sin_a * vec.y
        qy = sin_a * vec.x + cos_a * vec.y
        return Vector2D(qx, qy)

    @staticmethod
    def _reflect(vec, normal):
        """Hidden reflection method used by both Point2D and Vector2D."""
        d = 2 * (vec.x * normal.x + vec.y * normal.y)
        return Vector2D(vec.x - d * normal.x, vec.y - d * normal.y)

    def __copy__(self):
        return self.__class__(self.x, self.y)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.x, self.y)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, (Vector2D, Point2D)) and \
            self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __nonzero__(self):
        return self.x != 0 or self.y != 0

    def __len__(self):
        return 2

    def __getitem__(self, key):
        return (self.x, self.y)[key]

    def __iter__(self):
        return iter((self.x, self.y))

    def __add__(self, other):
        # Vector + Point -> Point
        # Vector + Vector -> Vector
        if isinstance(other, Point2D):
            return Point2D(self.x + other.x, self.y + other.y)
        elif isinstance(other, Vector2D):
            return Vector2D(self.x + other.x, self.y + other.y)
        else:
            raise TypeError('Cannot add {} and {}'.format(
                self.__class__.__name__, type(other)))

    __radd__ = __add__

    def __sub__(self, other):
        # Vector - Point -> Point
        # Vector - Vector -> Vector
        if isinstance(other, Point2D):
            return Point2D(self.x - other.x, self.y - other.y)
        elif isinstance(other, Vector2D):
            return Vector2D(self.x - other.x, self.y - other.y)
        else:
            raise TypeError('Cannot subtract {} and {}'.format(
                self.__class__.__name__, type(other)))

    def __rsub__(self, other):
        if isinstance(other, (Vector2D, Point2D)):
            return Vector2D(other.x - self.x, other.y - self.y)
        else:
            assert hasattr(other, '__len__') and len(other) == 2, \
                'Cannot subtract {} and {}'.format(
                    self.__class__.__name__, type(other))
            return Vector2D(other.x - self[0], other.y - self[1])

    def __mul__(self, other):
        assert type(other) in (int, float), \
            'Cannot multiply types {} and {}'.format(
                self.__class__.__name__, type(other))
        return Vector2D(self.x * other, self.y * other)

    __rmul__ = __mul__

    def __div__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector2D(self.x / other, self.y / other)

    def __rdiv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector2D(other / self.x, other / self.y)

    def __floordiv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector2D(operator.floordiv(self.x, other),
                        operator.floordiv(self.y, other))

    def __rfloordiv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector2D(operator.floordiv(other, self.x),
                        operator.floordiv(other, self.y))

    def __truediv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector2D(operator.truediv(self.x, other),
                        operator.truediv(self.y, other))

    def __rtruediv__(self, other):
        assert type(other) in (int, float), \
            'Cannot divide types {} and {}'.format(self.__class__.__name__, type(other))
        return Vector2D(operator.truediv(other, self.x),
                        operator.truediv(other, self.y))

    def __neg__(self):
        return Vector2D(-self.x, -self.y)

    __pos__ = __copy__

    def __abs__(self):
        return math.sqrt(self.x ** 2 + self.y ** 2)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Vector2D representation."""
        return 'Vector2D (%.2f, %.2f)' % (self.x, self.y)


class Point2D(Vector2D):
    """2D Point object.

    Args:
        x: Number for the X coordinate.
        y: Number for the Y coordinate.

    Properties:
        * x
        * y
    """
    __slots__ = ()

    def move(self, moving_vec):
        """Get a point that has been moved along a vector.

        Args:
            moving_vec: A Vector2D with the direction and distance to move the point.
        """
        return Point2D(self.x + moving_vec.x, self.y + moving_vec.y)

    def rotate(self, angle, origin):
        """Rotate a point counterclockwise by a certain angle around an origin.

        Args:
            angle: An angle for rotation in radians.
            origin: A Point2D for the origin around which the point will be rotated.
        """
        return Vector2D._rotate(self - origin, angle) + origin

    def reflect(self, normal, origin):
        """Get a point reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector2D representing the normal vector for the plane across
                which the point will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point2D representing the origin from which to reflect.
        """
        return Vector2D._reflect(self - origin, normal) + origin

    def scale(self, factor, origin=None):
        """Scale a point by a factor from an origin point.

        Args:
            factor: A number representing how much the point should be scaled.
            origin: A Point2D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0).
        """
        if origin is None:
            return Point2D(self.x * factor, self.y * factor)
        else:
            return (factor * (self - origin)) + origin

    def distance_to_point(self, point):
        """Get the distance from this point to another Point2D."""
        vec = (self.x - point.x, self.y - point.y)
        return math.sqrt(vec[0] ** 2 + vec[1] ** 2)

    def is_equivalent(self, point, tolerance):
        """Test whether this point is equivalent to another within a certain tolerance.

        Note that if you want to test whether the coordinate values are perfectly
        equal to one another, the == operator can be used.

        Args:
            point: Another Point2D for which geometric equivalency will be tested.
            tolerance: The minimum difference between the coordinate values of two
                points at which they can be considered geometrically equivalent.
        Returns:
            True if equivalent.  False if not equivalent.
        """
        return abs(self.x - point.x) <= tolerance and \
            abs(self.y - point.y) <= tolerance

    def to_dict(self):
        """Get Point2D as a dictionary."""
        return {'type': 'Point2D',
                'x': self.x,
                'y': self.y}

    def __add__(self, other):
        # Point + Vector -> Point
        # Point + Point -> Vector
        if isinstance(other, Point2D):
            return Vector2D(self.x + other.x, self.y + other.y)
        elif isinstance(other, Vector2D):
            return Point2D(self.x + other.x, self.y + other.y)
        else:
            raise TypeError('Cannot add Point2D and {}'.format(type(other)))

    def __sub__(self, other):
        # Point - Vector -> Point
        # Point - Point -> Vector
        if isinstance(other, Point2D):
            return Vector2D(self.x - other.x, self.y - other.y)
        elif isinstance(other, Vector2D):
            return Point2D(self.x - other.x, self.y - other.y)
        else:
            raise TypeError('Cannot subtract Point2D and {}'.format(type(other)))

    def __repr__(self):
        """Point2D representation."""
        return 'Point2D (%.2f, %.2f)' % (self.x, self.y)

    def __lt__(self, other):
        """ Lesser then inequality method. This is used by certain external
        data structure libraries to efficiently store and retrieve point data.
        """
        if isinstance(other, Vector2D):
            return self.x < other.x

    def __gt__(self, other):
        """ Greater then inequality method. This is used by certain external
        data structure libraries to efficiently store and retrieve point data.
        """
        if isinstance(other, Vector2D):
            return self.x > other.x




# coding=utf-8
import pytest

from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug_geometry.geometry3d.plane import Plane

import math


def test_linesegment3d_init():
    """Test the initalization of LineSegment3D objects and basic properties."""
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)
    str(seg)  # test the string representation of the line segment

    assert seg.p == Point3D(2, 0, 2)
    assert seg.v == Vector3D(0, 2, 0)
    assert seg.p1 == Point3D(2, 0, 2)
    assert seg.p2 == Point3D(2, 2, 2)
    assert seg.midpoint == Point3D(2, 1, 2)
    assert seg.point_at(0.25) == Point3D(2, 0.5, 2)
    assert seg.point_at_length(1) == Point3D(2, 1, 2)
    assert seg.length == 2

    flip_seg = seg.flip()
    assert flip_seg.p == Point3D(2, 2, 2)
    assert flip_seg.v == Vector3D(0, -2, 0)


def test_equality():
    """Test the equality of LineSegement3D objects."""
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)
    seg_dup = seg.duplicate()
    seg_alt = LineSegment3D(Point3D(2, 0.1, 2), vec)

    assert seg is seg
    assert seg is not seg_dup
    assert seg == seg_dup
    assert hash(seg) == hash(seg_dup)
    assert seg != seg_alt
    assert hash(seg) != hash(seg_alt)


def test_linesegment3_to_from_dict():
    """Test the to/from dict of LineSegment3D objects."""
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)
    seg_dict = seg.to_dict()
    new_seg = LineSegment3D.from_dict(seg_dict)
    assert isinstance(new_seg, LineSegment3D)
    assert new_seg.to_dict() == seg_dict


def test_init_from_endpoints():
    """Test the initalization of LineSegment3D from end points."""
    pt_1 = Point3D(2, 0, 2)
    pt_2 = Point3D(2, 2, 2)
    seg = LineSegment3D.from_end_points(pt_1, pt_2)

    assert seg.p == Point3D(2, 0, 2)
    assert seg.v == Vector3D(0, 2, 0)
    assert seg.p1 == Point3D(2, 0, 2)
    assert seg.p2 == Point3D(2, 2, 2)
    assert seg.length == 2


def test_init_from_sdl():
    """Test the initalization of LineSegment3D from start, direction, length."""
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 1, 0)
    seg = LineSegment3D.from_sdl(pt, vec, 2)

    assert seg.p == Point3D(2, 0, 2)
    assert seg.v == Vector3D(0, 2, 0)
    assert seg.p1 == Point3D(2, 0, 2)
    assert seg.p2 == Point3D(2, 2, 2)
    assert seg.length == 2


def test_linesegment3d_immutability():
    """Test the immutability of LineSegment3D objects."""
    pt = Point3D(2, 0, 0)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    assert isinstance(seg, LineSegment3D)
    with pytest.raises(AttributeError):
        seg.p.x = 3
    with pytest.raises(AttributeError):
        seg.v.x = 3
    with pytest.raises(AttributeError):
        seg.p = Point3D(0, 0, 0)
    with pytest.raises(AttributeError):
        seg.v = Vector3D(2, 2, 0)

    seg_copy = seg.duplicate()
    assert seg.p == seg_copy.p
    assert seg.v == seg_copy.v


def test_parallel_colinear():
    """Test the is_parallel and is_colinear methods."""
    pt_1 = Point3D(0, 0, 0)
    pt_2 = Point3D(0, 4, 0)
    pt_3 = Point3D(2, 0, 2)
    vec_1 = Vector3D(0, 2, 0)
    vec_2 = Vector3D(0, -2, 0)
    vec_3 = Vector3D(3, 3, 0)
    seg_1 = LineSegment3D(pt_1, vec_1)
    seg_2 = LineSegment3D(pt_2, vec_1)
    seg_3 = LineSegment3D(pt_1, vec_2)
    seg_4 = LineSegment3D(pt_3, vec_1)
    seg_5 = LineSegment3D(pt_3, vec_2)
    seg_6 = LineSegment3D(pt_1, vec_3)

    assert seg_1.is_colinear(seg_2, 0.0001, 0.0001)
    assert seg_1.is_colinear(seg_3, 0.0001, 0.0001)
    assert not seg_1.is_colinear(seg_4, 0.0001, 0.0001)

    assert seg_1.is_parallel(seg_4, 0.0001)
    assert seg_1.is_parallel(seg_5, 0.0001)
    assert not seg_1.is_parallel(seg_6, 0.0001)


def test_move():
    """Test the LineSegment3D move method."""
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    vec_1 = Vector3D(2, 2, 2)
    new_seg = seg.move(vec_1)
    assert new_seg.p == Point3D(4, 2, 4)
    assert new_seg.v == vec
    assert new_seg.p1 == Point3D(4, 2, 4)
    assert new_seg.p2 == Point3D(4, 4, 4)


def test_scale():
    """Test the LineSegment3D scale method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    origin_1 = Point3D(0, 2, 2)
    origin_2 = Point3D(1, 1, 2)
    new_seg = seg.scale(2, origin_1)
    assert new_seg.p == Point3D(4, 2, 2)
    assert new_seg.v == Point3D(0, 4, 0)
    assert new_seg.length == 4

    new_seg = seg.scale(2, origin_2)
    assert new_seg.p == Point3D(3, 3, 2)
    assert new_seg.v == Point3D(0, 4, 0)


def test_scale_world_origin():
    """Test the LineSegment3D scale method with None origin."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    new_seg = seg.scale(2)
    assert new_seg.p == Point3D(4, 4, 4)
    assert new_seg.v == Point3D(0, 4)
    assert new_seg.length == 4


def test_rotate():
    """Test the LineSegment3D rotate method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)
    origin_1 = Point3D(0, 0, 0)
    axis_1 = Vector3D(1, 0, 0)

    test_1 = seg.rotate(axis_1, math.pi, origin_1)
    assert test_1.p.x == pytest.approx(2, rel=1e-3)
    assert test_1.p.y == pytest.approx(-2, rel=1e-3)
    assert test_1.p.z == pytest.approx(-2, rel=1e-3)
    assert test_1.v.x == pytest.approx(0, rel=1e-3)
    assert test_1.v.y == pytest.approx(-2, rel=1e-3)
    assert test_1.v.z == pytest.approx(0, rel=1e-3)

    test_2 = seg.rotate(axis_1, math.pi/2, origin_1)
    assert test_2.p.x == pytest.approx(2, rel=1e-3)
    assert test_2.p.y == pytest.approx(-2, rel=1e-3)
    assert test_2.p.z == pytest.approx(2, rel=1e-3)
    assert test_2.v.x == pytest.approx(0, rel=1e-3)
    assert test_2.v.y == pytest.approx(0, rel=1e-3)
    assert test_2.v.z == pytest.approx(2, rel=1e-3)


def test_rotate_xy():
    """Test the LineSegment3D rotate_xy method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)
    origin_1 = Point3D(0, 2, 2)

    test_1 = seg.rotate_xy(math.pi, origin_1)
    assert test_1.p.x == pytest.approx(-2, rel=1e-3)
    assert test_1.p.y == pytest.approx(2, rel=1e-3)
    assert test_1.v.x == pytest.approx(0, rel=1e-3)
    assert test_1.v.y == pytest.approx(-2, rel=1e-3)

    test_2 = seg.rotate_xy(math.pi/2, origin_1)
    assert test_2.p.x == pytest.approx(0, rel=1e-3)
    assert test_2.p.y == pytest.approx(4, rel=1e-3)
    assert test_2.v.x == pytest.approx(-2, rel=1e-3)
    assert test_2.v.y == pytest.approx(0, rel=1e-3)


def test_reflect():
    """Test the LineSegment3D reflect method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    origin_1 = Point3D(0, 1, 2)
    origin_2 = Point3D(1, 1, 2)
    normal_1 = Vector3D(0, 1, 0)
    normal_2 = Vector3D(-1, 1, 0).normalize()

    assert seg.reflect(normal_1, origin_1).p == Point3D(2, 0, 2)
    assert seg.reflect(normal_1, origin_1).v == Vector3D(0, -2, 0)
    assert seg.reflect(normal_1, origin_2).p == Point3D(2, 0, 2)
    assert seg.reflect(normal_1, origin_2).v == Vector3D(0, -2, 0)

    test_1 = seg.reflect(normal_2, origin_2)
    assert test_1.p == Point3D(2, 2, 2)
    assert test_1.v.x == pytest.approx(2, rel=1e-3)
    assert test_1.v.y == pytest.approx(0, rel=1e-3)
    assert test_1.v.z == pytest.approx(0, rel=1e-3)

    test_2 = seg.reflect(normal_2, origin_1)
    assert test_2.p.x == pytest.approx(1, rel=1e-3)
    assert test_2.p.y == pytest.approx(3, rel=1e-3)
    assert test_2.p.z == pytest.approx(2, rel=1e-3)
    assert test_2.v.x == pytest.approx(2, rel=1e-3)
    assert test_2.v.y == pytest.approx(0, rel=1e-3)
    assert test_2.v.z == pytest.approx(0, rel=1e-3)


def test_subdivide():
    """Test the LineSegment3D subdivide methods."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    divisions = seg.subdivide(0.5)
    assert len(divisions) == 5
    assert divisions[0] == pt
    assert divisions[1] == Point3D(2, 2.5, 2)
    assert divisions[2] == Point3D(2, 3, 2)
    assert divisions[3] == Point3D(2, 3.5, 2)
    assert divisions[4] == Point3D(2, 4, 2)

    divisions = seg.subdivide([1, 0.5, 0.25])
    assert len(divisions) == 5
    assert divisions[0] == pt
    assert divisions[1] == Point3D(2, 3, 2)
    assert divisions[2] == Point3D(2, 3.5, 2)
    assert divisions[3] == Point3D(2, 3.75, 2)
    assert divisions[4] == Point3D(2, 4, 2)

    divisions = seg.subdivide_evenly(4)
    assert len(divisions) == 5
    assert divisions[0] == pt
    assert divisions[1] == Point3D(2, 2.5, 2)
    assert divisions[2] == Point3D(2, 3, 2)
    assert divisions[3] == Point3D(2, 3.5, 2)
    assert divisions[4] == Point3D(2, 4, 2)


def test_subdivide_tolerance_issue():
    """Test the LineSegment3D subdivide method handles Python tolerance.

    This case was taken from an originally failing situation in Rhino.
    """
    origin = Point3D(-68.773857131759087, 20.836167334772536, 18.897600000000001)
    dir = Vector3D(-0.026879795220556984, -0.99963867302585929, 0.0)
    bottom_seg = LineSegment3D.from_sdl(origin, dir, 21.5037894118)

    assert len(bottom_seg.subdivide_evenly(5)) == 6
    assert len(bottom_seg.subdivide_evenly(7)) == 8


def test_closest_point():
    """Test the LineSegment3D closest_point method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    near_pt = Point3D(3, 3, 0)
    assert seg.closest_point(near_pt) == Point3D(2, 3, 2)
    near_pt = Point3D(2, 0, 0)
    assert seg.closest_point(near_pt) == Point3D(2, 2, 2)
    near_pt = Point3D(2, 5, 0)
    assert seg.closest_point(near_pt) == Point3D(2, 4, 2)


def test_distance_to_point():
    """Test the LineSegment3D distance_to_point method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    near_pt = Point3D(3, 3, 2)
    assert seg.distance_to_point(near_pt) == 1
    near_pt = Point3D(2, 0, 2)
    assert seg.distance_to_point(near_pt) == 2
    near_pt = Point3D(2, 5, 2)
    assert seg.distance_to_point(near_pt) == 1


def test_intersect_plane():
    """Test the LineSegment3D intersect_plane method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    plane_1 = Plane(n=Vector3D(0, 1, 0), o=Point3D(2, 3, 2))
    assert isinstance(seg.intersect_plane(plane_1), Point3D)
    plane_2 = Plane(n=Vector3D(0, 1, 0), o=Point3D(2, 0, 2))
    assert seg.intersect_plane(plane_2) is None


def test_split_with_plane():
    """Test the LineSegment3D split_with_plane method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    seg = LineSegment3D(pt, vec)

    plane_1 = Plane(n=Vector3D(0, 1, 0), o=Point3D(2, 3, 2))
    assert len(seg.split_with_plane(plane_1)) == 2
    plane_2 = Plane(n=Vector3D(0, 1, 0), o=Point3D(2, 0, 2))
    assert len(seg.split_with_plane(plane_2)) == 1


def test_to_from_array():
    """Test to/from array method"""
    test_line = LineSegment3D.from_end_points(Point3D(2, 0, 2), Point3D(2, 2, 2))
    line_array = ((2, 0, 2), (2, 2, 2))

    assert test_line == LineSegment3D.from_array(line_array)

    line_array = ((2, 0, 2), (2, 2, 2))
    test_line = LineSegment3D.from_end_points(Point3D(2, 0, 2), Point3D(2, 2, 2))

    assert test_line.to_array() == line_array

    test_line_2 = LineSegment3D.from_array(test_line.to_array())
    assert test_line == test_line_2



# coding=utf-8
"""3D Ray"""
from __future__ import division

from .pointvector import Point3D, Vector3D
from ._1d import Base1DIn3D


class Ray3D(Base1DIn3D):
    """3D Ray object.

    Args:
        p: A Point3D representing the base of the ray.
        v: A Vector3D representing the direction of the ray.

    Properties:
        * p
        * v
    """
    __slots__ = ()

    def __init__(self, p, v):
        """Initialize Ray3D."""
        Base1DIn3D.__init__(self, p, v)
    
    @classmethod
    def from_array(cls, ray_array):
        """ Create a Ray3D from a nested array with a point and a vector.

        Args:
            ray_array: Nested tuples ((p.x, p.y), (v.x, v.y)).
        """
        return Ray3D(Point3D(*ray_array[0]), Vector3D(*ray_array[1]))

    def reverse(self):
        """Get a copy of this ray that is reversed."""
        return Ray3D(self.p, self.v.reverse())

    def move(self, moving_vec):
        """Get a ray that has been moved along a vector.

        Args:
            moving_vec: A Vector3D with the direction and distance to move the ray.
        """
        return Ray3D(self.p.move(moving_vec), self.v)

    def rotate(self, axis, angle, origin):
        """Rotate a ray by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis: A Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in radians.
            origin: A Point3D for the origin around which the object will be rotated.
        """
        return Ray3D(self.p.rotate(axis, angle, origin), self.v.rotate(axis, angle))

    def rotate_xy(self, angle, origin):
        """Get a ray rotated counterclockwise in the XY plane by a certain angle.

        Args:
            angle: An angle in radians.
            origin: A Point3D for the origin around which the object will be rotated.
        """
        return Ray3D(self.p.rotate_xy(angle, origin), self.v.rotate_xy(angle))

    def reflect(self, normal, origin):
        """Get a ray reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector3D representing the normal vector for the plane across
                which the ray will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point3D representing the origin from which to reflect.
        """
        return Ray3D(self.p.reflect(normal, origin), self.v.reflect(normal))

    def scale(self, factor, origin=None):
        """Scale a ray by a factor from an origin point.

        Args:
            factor: A number representing how much the ray should be scaled.
            origin: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        return Ray3D(self.p.scale(factor, origin), self.v * factor)

    def scale_world_origin(self, factor):
        """Scale a ray by a factor from the world origin. Faster than Ray2D.scale.

        Args:
            factor: A number representing how much the ray should be scaled.
        """
        return Ray3D(self.p.scale_world_origin(factor), self.v * factor)

    def to_dict(self):
        """Get Ray3D as a dictionary."""
        base = Base1DIn3D.to_dict(self)
        base['type'] = 'Ray3D'
        return base

    def to_array(self):
        """A nested array representing the start point and vector."""
        return (self.p.to_array(), self.v.to_array())

    def _u_in(self, u):
        return u >= 0.0
    
    def __copy__(self):
        return Ray3D(self.p, self.v)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self.p), hash(self.v))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Ray3D) and self.__key() == other.__key()

    def __repr__(self):
        return 'Ray3D (point <%.2f, %.2f, %.2f>) (vector <%.2f, %.2f, %.2f>)' % \
            (self.p.x, self.p.y, self.p.z, self.v.x, self.v.y, self.v.z)



# coding=utf-8
import pytest

from ladybug_geometry.geometry2d.polyline import Polyline2D
from ladybug_geometry.geometry2d.polygon import Polygon2D
from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.ray import Ray2D

import math


def test_polyline2d_init():
    """Test the initalization of Polyline2D objects and basic properties."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = Polyline2D(pts)

    str(pline)  # test the string representation of the polyline

    assert isinstance(pline.vertices, tuple)
    assert len(pline.vertices) == 4
    assert len(pline) == 4
    for point in pline:
        assert isinstance(point, Point2D)

    assert isinstance(pline.segments, tuple)
    assert len(pline.segments) == 3
    for seg in pline.segments:
        assert isinstance(seg, LineSegment2D)
        assert seg.length == 2

    assert pline.length == 6
    assert pline.is_self_intersecting is False

    assert pline.vertices[0] == pline[0]

    p_array = pline.to_array()
    assert isinstance(p_array, tuple)
    assert len(p_array) == 4
    for arr in p_array:
        assert isinstance(p_array, tuple)
        assert len(arr) == 2
    pline_2 = Polyline2D.from_array(p_array)
    assert pline == pline_2


def test_equality():
    """Test the equality of Polyline2D objects."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pts_2 = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0.1, 2))
    pline = Polyline2D(pts)
    pline_dup = pline.duplicate()
    pline_alt = Polyline2D(pts_2)

    assert pline is pline
    assert pline is not pline_dup
    assert pline == pline_dup
    assert hash(pline) == hash(pline_dup)
    assert pline != pline_alt
    assert hash(pline) != hash(pline_alt)


def test_to_polygon():
    """Test the to_polygon method."""
    pts_1 = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pts_2 = pts_1 + (Point2D(0, 0),)

    pline_1 = Polyline2D(pts_1)
    pline_2 = Polyline2D(pts_2)

    polygon_1 = pline_1.to_polygon(0.01)
    polygon_2 = pline_2.to_polygon(0.01)

    assert isinstance(polygon_1, Polygon2D)
    assert isinstance(polygon_2, Polygon2D)
    assert len(polygon_1) == 4
    assert len(polygon_2) == 4


def test_from_polygon():
    """Test the from_polygon method."""
    pts_1 = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pgon = Polygon2D(pts_1)

    pline = Polyline2D.from_polygon(pgon)
    assert isinstance(pline, Polyline2D)
    assert len(pline) == 5


def test_polyline2d_to_from_dict():
    """Test the to/from dict of Polyline2D objects."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = Polyline2D(pts)
    pline_dict = pline.to_dict()
    new_pline = Polyline2D.from_dict(pline_dict)
    assert isinstance(new_pline, Polyline2D)
    assert new_pline.to_dict() == pline_dict


def test_is_self_intersecting():
    """Test the is_self_intersecting property."""
    pts_1 = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline_1 = Polyline2D(pts_1)
    pts_2 = (Point2D(0, 0), Point2D(0, 2), Point2D(2, 0), Point2D(2, 2),
             Point2D(0.5, 0.5))
    pline_2 = Polyline2D(pts_2)

    assert not pline_1.is_self_intersecting
    assert pline_2.is_self_intersecting


def test_min_max_center():
    """Test the Polyline2D min, max and center."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = Polyline2D(pts)

    assert pline.min == Point2D(0, 0)
    assert pline.max == Point2D(2, 2)
    assert pline.center == Point2D(1, 1)


def test_remove_colinear_vertices():
    """Test the remove_colinear_vertices method of Polyline2D."""
    pts_1 = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pts_2 = (Point2D(0, 0), Point2D(1, 0), Point2D(2, 0), Point2D(2, 2),
             Point2D(0, 2))
    pline_1 = Polyline2D(pts_1)
    pline_2 = Polyline2D(pts_2)

    assert len(pline_1.remove_colinear_vertices(0.0001).vertices) == 4
    assert len(pline_2.remove_colinear_vertices(0.0001).vertices) == 4


def test_pline2d_duplicate():
    """Test the duplicate method of Polyline2D."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = Polyline2D(pts)
    new_pline = pline.duplicate()

    for i, pt in enumerate(new_pline):
        assert pt == pts[i]

    assert pline.length == new_pline.length
    assert pline.vertices == new_pline.vertices
    assert pline.is_self_intersecting == new_pline.is_self_intersecting


def test_reverse():
    """Test the reverse method."""
    pts_1 = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = Polyline2D(pts_1)
    new_pline = pline.reverse()

    assert pline.length == new_pline.length
    assert pline.vertices == tuple(reversed(new_pline.vertices))
    assert pline.is_self_intersecting == new_pline.is_self_intersecting


def test_move():
    """Test the Polyline2D move method."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = Polyline2D(pts)

    vec_1 = Vector2D(2, 2)
    new_pline = pline.move(vec_1)
    assert new_pline[0] == Point2D(2, 2)
    assert new_pline[1] == Point2D(4, 2)
    assert new_pline[2] == Point2D(4, 4)
    assert new_pline[3] == Point2D(2, 4)

    assert pline.length == new_pline.length
    assert pline.is_self_intersecting is new_pline.is_self_intersecting


def test_scale():
    """Test the Polyline2D scale method."""
    pts_1 = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline_1 = Polyline2D(pts_1)
    pts_2 = (Point2D(1, 1), Point2D(2, 1), Point2D(2, 2), Point2D(1, 2))
    pline_2 = Polyline2D(pts_2)
    origin_1 = Point2D(2, 0)
    origin_2 = Point2D(1, 1)

    new_pline_1 = pline_1.scale(2, origin_1)
    assert new_pline_1[0] == Point2D(-2, 0)
    assert new_pline_1[1] == Point2D(2, 0)
    assert new_pline_1[2] == Point2D(2, 4)
    assert new_pline_1[3] == Point2D(-2, 4)
    assert new_pline_1.length == pline_1.length * 2

    new_pline_2 = pline_2.scale(2, origin_2)
    assert new_pline_2[0] == Point2D(1, 1)
    assert new_pline_2[1] == Point2D(3, 1)
    assert new_pline_2[2] == Point2D(3, 3)
    assert new_pline_2[3] == Point2D(1, 3)
    assert new_pline_2.length == pline_2.length * 2


def test_scale_world_origin():
    """Test the Polyline2D scale method with None origin."""
    pts = (Point2D(1, 1), Point2D(2, 1), Point2D(2, 2), Point2D(1, 2))
    pline = Polyline2D(pts)

    new_pline = pline.scale(2)
    assert new_pline[0] == Point2D(2, 2)
    assert new_pline[1] == Point2D(4, 2)
    assert new_pline[2] == Point2D(4, 4)
    assert new_pline[3] == Point2D(2, 4)
    assert new_pline.length == pline.length * 2


def test_rotate():
    """Test the Polyline2D rotate method."""
    pts = (Point2D(1, 1), Point2D(2, 1), Point2D(2, 2), Point2D(1, 2))
    pline = Polyline2D(pts)
    origin_1 = Point2D(1, 1)

    test_1 = pline.rotate(math.pi, origin_1)
    assert test_1[0].x == pytest.approx(1, rel=1e-3)
    assert test_1[0].y == pytest.approx(1, rel=1e-3)
    assert test_1[2].x == pytest.approx(0, rel=1e-3)
    assert test_1[2].y == pytest.approx(0, rel=1e-3)
    assert pline.length == pytest.approx(test_1.length, rel=1e-3)

    test_2 = pline.rotate(math.pi/2, origin_1)
    assert test_2[0].x == pytest.approx(1, rel=1e-3)
    assert test_2[0].y == pytest.approx(1, rel=1e-3)
    assert test_2[2].x == pytest.approx(0, rel=1e-3)
    assert test_2[2].y == pytest.approx(2, rel=1e-3)


def test_reflect():
    """Test the Polyline2D reflect method."""
    pts = (Point2D(1, 1), Point2D(2, 1), Point2D(2, 2), Point2D(1, 2))
    pline = Polyline2D(pts)

    origin_1 = Point2D(1, 0)
    normal_1 = Vector2D(1, 0)
    normal_2 = Vector2D(-1, -1).normalize()

    test_1 = pline.reflect(normal_1, origin_1)
    assert test_1[0].x == pytest.approx(1, rel=1e-3)
    assert test_1[0].y == pytest.approx(1, rel=1e-3)
    assert test_1[2].x == pytest.approx(0, rel=1e-3)
    assert test_1[2].y == pytest.approx(2, rel=1e-3)
    assert pline.length == pytest.approx(test_1.length, rel=1e-3)


def test_intersect_line_ray():
    """Test the Polyline2D intersect_line_ray method."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = Polyline2D(pts)

    ray_1 = Ray2D(Point2D(1, -1), Vector2D(0, 1))
    ray_2 = Ray2D(Point2D(1, 1), Vector2D(1, 0))
    ray_3 = Ray2D(Point2D(1, 1), Vector2D(11, 0))
    ray_4 = Ray2D(Point2D(-1, 1), Vector2D(-1, 0))
    ray_5 = Ray2D(Point2D(-1, 1), Vector2D(1, 0))

    assert len(pline.intersect_line_ray(ray_1)) == 2
    assert len(pline.intersect_line_ray(ray_2)) == 1
    assert len(pline.intersect_line_ray(ray_3)) == 1
    assert len(pline.intersect_line_ray(ray_4)) == 0
    assert len(pline.intersect_line_ray(ray_5)) == 1

    line_1 = LineSegment2D(Point2D(-1, 1), Vector2D(0.5, 0))
    line_2 = LineSegment2D(Point2D(1, -1), Vector2D(0, 2))
    line_3 = LineSegment2D(Point2D(1, -1), Vector2D(0, 3))

    assert len(pline.intersect_line_ray(line_1)) == 0
    assert len(pline.intersect_line_ray(line_2)) == 1
    assert len(pline.intersect_line_ray(line_3)) == 2


def test_intersect_line_infinite():
    """Test the Polyline2D intersect_line_infinite method."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    pline = Polyline2D(pts)

    ray_1 = Ray2D(Point2D(-1, 1), Vector2D(1, 0))
    ray_2 = Ray2D(Point2D(1, 1), Vector2D(1, 0))
    ray_3 = Ray2D(Point2D(1, 1), Vector2D(11, 0))
    ray_4 = Ray2D(Point2D(-1, 1), Vector2D(-1, 0))
    ray_5 = Ray2D(Point2D(-1, 3), Vector2D(-1, 0))
    ray_6 = Ray2D(Point2D(0, 2), Vector2D(-1, -1))

    assert len(pline.intersect_line_infinite(ray_1)) == 1
    assert len(pline.intersect_line_infinite(ray_2)) == 1
    assert len(pline.intersect_line_infinite(ray_3)) == 1
    assert len(pline.intersect_line_infinite(ray_4)) == 1
    assert len(pline.intersect_line_infinite(ray_5)) == 0
    assert len(pline.intersect_line_infinite(ray_6)) > 0

def test_join_segments():
    """Test the join_segments method."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    l_segs = (LineSegment2D.from_end_points(pts[0], pts[1]),
            LineSegment2D.from_end_points(pts[1], pts[2]),
            LineSegment2D.from_end_points(pts[2], pts[3]),
            LineSegment2D.from_end_points(pts[3], pts[0]))
    p_lines = Polyline2D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 1
    assert isinstance(p_lines[0], Polyline2D)
    assert len(p_lines[0]) == 5
    assert p_lines[0].is_closed(0.01)

    l_segs = (LineSegment2D.from_end_points(pts[0], pts[1]),
            LineSegment2D.from_end_points(pts[2], pts[3]),
            LineSegment2D.from_end_points(pts[1], pts[2]),
            LineSegment2D.from_end_points(pts[3], pts[0]))
    p_lines = Polyline2D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 1
    assert isinstance(p_lines[0], Polyline2D)
    assert len(p_lines[0]) == 5
    assert p_lines[0].is_closed(0.01)

    l_segs = (LineSegment2D.from_end_points(pts[0], pts[1]),
            LineSegment2D.from_end_points(pts[1], pts[2]),
            LineSegment2D.from_end_points(pts[0], pts[3]),
            LineSegment2D.from_end_points(pts[3], pts[2]))
    p_lines = Polyline2D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 1
    assert isinstance(p_lines[0], Polyline2D)
    assert len(p_lines[0]) == 5
    assert p_lines[0].is_closed(0.01)


def test_join_segments_multiple_pline():
    """Test the join_segments method with multiple polylines."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    extra_pts = (Point2D(3, 3), Point2D(4, 3), Point2D(4, 4), Point2D(3, 4))
    l_segs = (LineSegment2D.from_end_points(extra_pts[0], extra_pts[1]),
              LineSegment2D.from_end_points(pts[0], pts[1]),
              LineSegment2D.from_end_points(pts[1], pts[2]),
              LineSegment2D.from_end_points(pts[0], pts[3]),
              LineSegment2D.from_end_points(pts[3], pts[2]))
    p_lines = Polyline2D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 2

    l_segs = (LineSegment2D.from_end_points(extra_pts[0], extra_pts[1]),
              LineSegment2D.from_end_points(pts[0], pts[1]),
              LineSegment2D.from_end_points(pts[1], pts[2]),
              LineSegment2D.from_end_points(pts[0], pts[3]),
              LineSegment2D.from_end_points(pts[3], pts[2]),
              LineSegment2D.from_end_points(extra_pts[2], extra_pts[3]),
              LineSegment2D.from_end_points(extra_pts[1], extra_pts[2]),
              LineSegment2D.from_end_points(extra_pts[0], extra_pts[3]))
    p_lines = Polyline2D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 2
    for p_line in p_lines:
        assert isinstance(p_line, Polyline2D)
        assert len(p_line) == 5
        assert p_line.is_closed(0.01)


def test_join_segments_disconnected():
    """Test the join_segments method with diconnected polylines."""
    pts = (Point2D(0, 0), Point2D(2, 0), Point2D(2, 2), Point2D(0, 2))
    extra_pts = (Point2D(3, 3), Point2D(4, 3), Point2D(4, 4), Point2D(3, 4))

    l_segs = (LineSegment2D.from_end_points(extra_pts[0], extra_pts[1]),
              LineSegment2D.from_end_points(pts[0], pts[1]),
              LineSegment2D.from_end_points(pts[2], pts[3]),
              LineSegment2D.from_end_points(extra_pts[3], extra_pts[2]))
    p_lines = Polyline2D.join_segments(l_segs, 0.01)
    assert len(p_lines) == 4



# coding=utf-8
"""Cylinder"""
from __future__ import division

from .pointvector import Point3D, Vector3D

import math


class Cylinder(object):
    """Cylinder object.

    Args:
        center: A Point3D at the center of the bottom base of the cylinder.
        axis: A Vector3D representing the direction and height of the cylinder.
            The vector extends from the bottom base center to the top base center.
        radius: A number representing the radius of the cylinder.

    Properties:
        * center
        * axis
        * radius
        * center_end
        * diameter
        * height
        * area
        * volume
    """
    __slots__ = ('_center', '_axis', '_radius')

    def __init__(self, center, axis, radius):
        """Initialize Cylinder."""
        assert isinstance(center, Point3D), \
            "Expected Point3D. Got {}.".format(type(center))
        assert isinstance(axis, Vector3D), \
            "Expected Vector3D. Got {}.".format(type(axis))
        assert radius > 0, \
            'Cylinder radius must be greater than 0. Got {}.'.format(radius)
        self._center = center
        self._axis = axis
        self._radius = radius

    @classmethod
    def from_start_end(cls, p1, p2, radius):
        """Initialize a new cylinder from start and end points.

        Args:
            p1: The start point of the cylinder, represents the center of the
                bottom base of the cylinder.
            p2: The end point of the cylinder, represents the center of the top
                base of the cylinder
            radius: A number representing the radius of the cylinder.
        """
        axis = p2 - p1
        return cls(p1, axis, radius)

    @classmethod
    def from_dict(cls, data):
        """Create a Cylinder from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "type": "Cylinder"
            "center": (10, 0, 0),
            "axis": (0, 0, 1),
            "radius": 1.0,
            }
        """
        return cls(Point3D.from_array(data['center']),
                   Vector3D.from_array(data['axis']),
                   data['radius'])

    @property
    def center(self):
        """Center of Cylinder."""
        return self._center

    @property
    def axis(self):
        """Axis of Cylinder."""
        return self._axis

    @property
    def radius(self):
        """Radius of Cylinder"""
        return self._radius

    @property
    def center_end(self):
        """Center of the opposite end of Cylinder."""
        return self.center + self.axis

    @property
    def diameter(self):
        """Diameter of Cylinder"""
        return self.radius * 2

    @property
    def height(self):
        """Height of Cylinder"""
        return self.axis.magnitude

    @property
    def area(self):
        """Surface area of a Cylinder"""
        return 2 * math.pi * self.radius * self.height + 2 * math.pi * self.radius ** 2

    @property
    def volume(self):
        """Volume of a Cylinder"""
        return math.pi * self.radius ** 2 * self.height

    def move(self, moving_vec):
        """Get a Cylinder that has been moved along a vector.

        Args:
            moving_vec: A Vector3D with the direction and distance to move the Cylinder.
        """
        return Cylinder(self.center.move(moving_vec), self.axis, self.radius)

    def rotate(self, axis, angle, origin):
        """Rotate this Cylinder by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis: A Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in radians.
            origin: A Point3D for the origin around which the cylinder will be rotated.
        """
        return Cylinder(self.center.rotate(axis, angle, origin),
                        self.axis.rotate(axis, angle),
                        self.radius)

    def rotate_xy(self, angle, origin):
        """Get a Cylinder that is rotated counterclockwise in the world XY plane by an angle.

        Args:
            angle: An angle for rotation in radians.
            origin: A Point3D for the origin around which the cylinder will be rotated.
        """
        return Cylinder(self.center.rotate_xy(angle, origin),
                        self.axis.rotate_xy(angle),
                        self.radius)

    def reflect(self, normal, origin):
        """Get a Cylinder reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector3D representing the normal vector for the plane across
                which the arc will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point3D representing the origin from which to reflect.
        """
        return Cylinder(self.center.reflect(normal, origin),
                        self.axis.reflect(normal),
                        self.radius)

    def scale(self, factor, origin=None):
        """Scale a Cylinder by a factor from an origin point.

        Args:
            factor: A number representing how much the Cylinder should be scaled.
            origin: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        return Cylinder(self.center.scale(factor, origin),
                        self.axis * factor,
                        self.radius * factor)

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def to_dict(self):
        """Get Cylinder as a dictionary."""
        return {'type': 'Cylinder', 'center': self.center.to_array(),
                'axis': self.axis.to_array(), 'radius': self.radius}

    def __copy__(self):
        return Cylinder(self.center, self.axis, self.radius)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self._center, self._axis, self._radius)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Cylinder) and self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        return 'Cylinder (center {}) (axis {}) (radius {})'.\
            format(self.center, self.axis, self.radius)



# coding=utf-8
"""2D Line Segment"""
from __future__ import division

from .pointvector import Vector2D, Point2D
from ._1d import Base1DIn2D
from ..intersection2d import closest_point2d_between_line2d


class LineSegment2D(Base1DIn2D):
    """2D line segment object.

    Args:
        p: A Point2D representing the first point of the line segment.
        v: A Vector2D representing the vector to the second point.

    Properties:
        * p
        * v
        * p1
        * p2
        * midpoint
        * length
    """
    __slots__ = ()

    def __init__(self, p, v):
        """Initialize LineSegment2D."""
        Base1DIn2D.__init__(self, p, v)

    @classmethod
    def from_end_points(cls, p1, p2):
        """Initialize a line segment from a start point and and end point.

        Args:
            p1: A Point2D representing the first point of the line segment.
            p2: A Point2D representing the second point of the line segment.
        """
        v = p2 - p1
        return cls(p1, Vector2D(v.x, v.y))

    @classmethod
    def from_sdl(cls, s, d, length):
        """Initialize a line segment from a start point, direction, and length.

        Args:
            s: A Point2D representing the start point of the line segment.
            d: A Vector2D representing the direction of the line segment.
            length: A number representing the length of the line segment.
        """
        return cls(s, d * length / d.magnitude)

    @classmethod
    def from_array(cls, line_array):
        """ Create a LineSegment2D from a nested array of two endpoint coordinates.

        Args:
            line_array: Nested tuples ((pt1.x, pt1.y), (pt2.x, pt2.y)), where
                pt1 and pt2 represent the endpoints of the line segment.
        """
        return LineSegment2D.from_end_points(*tuple(Point2D(*pt) for pt in line_array))

    @property
    def p1(self):
        """First point (same as p)."""
        return self.p

    @property
    def p2(self):
        """Second point."""
        return Point2D(self.p.x + self.v.x, self.p.y + self.v.y)

    @property
    def endpoints(self):
        """ Tuple of endpoints """
        return (self.p1, self.p2)

    @property
    def midpoint(self):
        """Midpoint."""
        return self.point_at(0.5)

    @property
    def length(self):
        """The length of the line segment."""
        return self.v.magnitude

    def is_equivalent(self, other, tolerance):
        """Boolean noting equivalence (within tolerance) between this line and another.

        The order of the line points do not matter for equivalence to be true.

        Args:
            other: LineSegment2D for comparison.
            tolerance: float representing point equivalence.

        Returns:
            True if equivalent else False
        """
        return (self.p1.is_equivalent(other.p1, tolerance) and
                self.p2.is_equivalent(other.p2, tolerance)) or \
               (self.p1.is_equivalent(other.p2, tolerance) and
                self.p2.is_equivalent(other.p1, tolerance))

    def flip(self):
        """Get a copy of this line segment that is flipped."""
        return LineSegment2D(self.p2, self.v.reverse())

    def move(self, moving_vec):
        """Get a line segment that has been moved along a vector.

        Args:
            moving_vec: A Vector2D with the direction and distance to move the ray.
        """
        return LineSegment2D(self.p.move(moving_vec), self.v)

    def rotate(self, angle, origin):
        """Get a line segment that is rotated counterclockwise by a certain angle.

        Args:
            angle: An angle for rotation in radians.
            origin: A Point2D for the origin around which the line segment will
                be rotated.
        """
        return LineSegment2D(self.p.rotate(angle, origin), self.v.rotate(angle))

    def reflect(self, normal, origin):
        """Get a line segment reflected across a plane with the input normal vector and origin.

        Args:
            normal: A Vector2D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin: A Point2D representing the origin from which to reflect.
        """
        return LineSegment2D(self.p.reflect(normal, origin), self.v.reflect(normal))

    def scale(self, factor, origin=None):
        """Scale a line segment by a factor from an origin point.

        Args:
            factor: A number representing how much the line segment should be scaled.
            origin: A Point2D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0).
        """
        return LineSegment2D(self.p.scale(factor, origin), self.v * factor)

    def subdivide(self, distances):
        """Get Point2D values along the line that subdivide it based on input distances.

        Args:
            distances: A list of distances along the line at which to subdivide it.
                This can also be a single number that will be repeated until the
                end of the line.
        """
        if isinstance(distances, (float, int)):
            distances = [distances]
        line_length = self.length
        dist = distances[0]
        index = 0
        sub_pts = [self.p]
        while dist < line_length:
            sub_pts.append(self.point_at_length(dist))
            if index < len(distances) - 1:
                index += 1
            dist += distances[index]
        sub_pts.append(self.p2)
        return sub_pts

    def subdivide_evenly(self, number):
        """Get Point2D values along the line that divide it into evenly-spaced segments.

        Args:
            number: The number of segments into which the line will be divided.
        """
        interval = 1 / number
        parameter = interval
        sub_pts = [self.p]
        while parameter <= 1:
            sub_pts.append(self.point_at(parameter))
            parameter += interval
        return sub_pts

    def point_at(self, parameter):
        """Get a point at a given fraction along the line segment.

        Args:
            parameter: The fraction between the start and end point where the
                desired point lies. For example, 0.5 will yield the midpoint.
        """
        return self.p + self.v * parameter

    def point_at_length(self, length):
        """Get a point at a given distance along the line segment.

        Args:
            length: The distance along the line from the start point where the
                desired point lies.
        """
        return self.p + self.v * (length / self.length)

    def closest_points_between_line(self, line):
        """Get the two closest Point2D between this object to another LineSegment2D.

        Note that the line segments should not intersect for the result to be valid.

        Args:
            line: A LineSegment2D object to which the closest points
                will be computed.

        Returns:
            Two Point2D objects representing

            1) The closest point on this object to the input line.
            2) The closest point on the input line to this object.
        """
        dist, pts = closest_point2d_between_line2d(self, line)
        return pts

    def distance_to_line(self, line):
        """Get the minimum distance between this object and the input LineSegment2D.

        Note that the line segments should not intersect for the result to be valid.

        Args:
            line: A LineSegment2D object to which the minimum distance will be computed.

        Returns:
            The minimum distance to the input line.
        """
        dist, pts = closest_point2d_between_line2d(self, line)
        return dist

    def to_dict(self):
        """Get LineSegment2D as a dictionary."""
        base = Base1DIn2D.to_dict(self)
        base['type'] = 'LineSegment2D'
        return base

    def to_array(self):
        """ A nested list representing the two line endpoint coordinates."""
        return (self.p1.to_array(), self.p2.to_array())

    def _u_in(self, u):
        return u >= 0.0 and u <= 1.0

    def __abs__(self):
        return abs(self.v)

    def __copy__(self):
        return LineSegment2D(self.p, self.v)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self.p), hash(self.v))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, LineSegment2D) and self.__key() == other.__key()

    def __repr__(self):
        return 'LineSegment2D (<%.2f, %.2f> to <%.2f, %.2f>)' % \
            (self.p.x, self.p.y, self.p.x + self.v.x, self.p.y + self.v.y)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ladybug-geometry",
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author="Ladybug Tools",
    author_email="info@ladybug.tools",
    description="Ladybug geometry is a Python library that houses all of the basic geometry computation needed for Ladybug Tools core libraries.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ladybug-tools/ladybug-geometry",
    packages=setuptools.find_packages(exclude=["tests"]),
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent"
    ],
)



# coding=utf-8
import pytest

from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.ray import Ray3D

import math


def test_ray3d_init():
    """Test the initalization of Ray3D objects and basic properties."""
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)
    str(ray)  # test the string representation of the ray

    assert ray.p == Point3D(2, 0, 2)
    assert ray.v == Vector3D(0, 2, 0)

    flip_ray = ray.reverse()
    assert flip_ray.p == Point3D(2, 0, 2)
    assert flip_ray.v == Vector3D(0, -2, 0)


def test_equality():
    """Test the equality of Ray3D objects."""
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)
    ray_dup = ray.duplicate()
    ray_alt = Ray3D(Point3D(2, 0.1, 2), vec)

    assert ray is ray
    assert ray is not ray_dup
    assert ray == ray_dup
    assert hash(ray) == hash(ray_dup)
    assert ray != ray_alt
    assert hash(ray) != hash(ray_alt)


def test_ray3_to_from_dict():
    """Test the to/from dict of Ray3D objects."""
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)
    ray_dict = ray.to_dict()
    new_ray = Ray3D.from_dict(ray_dict)
    assert isinstance(new_ray, Ray3D)
    assert new_ray.to_dict() == ray_dict


def test_ray3d_immutability():
    """Test the immutability of Ray3D objects."""
    pt = Point3D(2, 0, 0)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)

    assert isinstance(ray, Ray3D)
    with pytest.raises(AttributeError):
        ray.p.x = 3
    with pytest.raises(AttributeError):
        ray.v.x = 3
    with pytest.raises(AttributeError):
        ray.p = Point3D(0, 0, 0)
    with pytest.raises(AttributeError):
        ray.v = Vector3D(2, 2, 0)

    ray_copy = ray.duplicate()
    assert ray.p == ray_copy.p
    assert ray.v == ray_copy.v


def test_move():
    """Test the Ray3D move method."""
    pt = Point3D(2, 0, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)

    vec_1 = Vector3D(2, 2, 2)
    new_ray = ray.move(vec_1)
    assert new_ray.p == Point3D(4, 2, 4)
    assert new_ray.v == vec
    assert new_ray.v.magnitude == 2


def test_scale():
    """Test the Ray3D scale method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)

    origin_1 = Point3D(0, 2, 2)
    origin_2 = Point3D(1, 1, 2)
    new_ray = ray.scale(2, origin_1)
    assert new_ray.p == Point3D(4, 2, 2)
    assert new_ray.v == Point3D(0, 4, 0)
    assert new_ray.v.magnitude == 4

    new_ray = ray.scale(2, origin_2)
    assert new_ray.p == Point3D(3, 3, 2)
    assert new_ray.v == Point3D(0, 4, 0)


def test_scale_world_origin():
    """Test the Ray3D scale method with None origin."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)

    new_ray = ray.scale(2)
    assert new_ray.p == Point3D(4, 4, 4)
    assert new_ray.v == Point3D(0, 4)
    assert new_ray.v.magnitude == 4


def test_rotate():
    """Test the Ray3D rotate method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)
    origin_1 = Point3D(0, 0, 0)
    axis_1 = Vector3D(1, 0, 0)

    test_1 = ray.rotate(axis_1, math.pi, origin_1)
    assert test_1.p.x == pytest.approx(2, rel=1e-3)
    assert test_1.p.y == pytest.approx(-2, rel=1e-3)
    assert test_1.p.z == pytest.approx(-2, rel=1e-3)
    assert test_1.v.x == pytest.approx(0, rel=1e-3)
    assert test_1.v.y == pytest.approx(-2, rel=1e-3)
    assert test_1.v.z == pytest.approx(0, rel=1e-3)

    test_2 = ray.rotate(axis_1, math.pi/2, origin_1)
    assert test_2.p.x == pytest.approx(2, rel=1e-3)
    assert test_2.p.y == pytest.approx(-2, rel=1e-3)
    assert test_2.p.z == pytest.approx(2, rel=1e-3)
    assert test_2.v.x == pytest.approx(0, rel=1e-3)
    assert test_2.v.y == pytest.approx(0, rel=1e-3)
    assert test_2.v.z == pytest.approx(2, rel=1e-3)


def test_rotate_xy():
    """Test the Ray3D rotate_xy method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)
    origin_1 = Point3D(0, 2, 2)

    test_1 = ray.rotate_xy(math.pi, origin_1)
    assert test_1.p.x == pytest.approx(-2, rel=1e-3)
    assert test_1.p.y == pytest.approx(2, rel=1e-3)
    assert test_1.v.x == pytest.approx(0, rel=1e-3)
    assert test_1.v.y == pytest.approx(-2, rel=1e-3)

    test_2 = ray.rotate_xy(math.pi/2, origin_1)
    assert test_2.p.x == pytest.approx(0, rel=1e-3)
    assert test_2.p.y == pytest.approx(4, rel=1e-3)
    assert test_2.v.x == pytest.approx(-2, rel=1e-3)
    assert test_2.v.y == pytest.approx(0, rel=1e-3)


def test_reflect():
    """Test the Ray3D reflect method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)

    origin_1 = Point3D(0, 1, 2)
    origin_2 = Point3D(1, 1, 2)
    normal_1 = Vector3D(0, 1, 0)
    normal_2 = Vector3D(-1, 1, 0).normalize()

    assert ray.reflect(normal_1, origin_1).p == Point3D(2, 0, 2)
    assert ray.reflect(normal_1, origin_1).v == Vector3D(0, -2, 0)
    assert ray.reflect(normal_1, origin_2).p == Point3D(2, 0, 2)
    assert ray.reflect(normal_1, origin_2).v == Vector3D(0, -2, 0)

    test_1 = ray.reflect(normal_2, origin_2)
    assert test_1.p == Point3D(2, 2, 2)
    assert test_1.v.x == pytest.approx(2, rel=1e-3)
    assert test_1.v.y == pytest.approx(0, rel=1e-3)
    assert test_1.v.z == pytest.approx(0, rel=1e-3)

    test_2 = ray.reflect(normal_2, origin_1)
    assert test_2.p.x == pytest.approx(1, rel=1e-3)
    assert test_2.p.y == pytest.approx(3, rel=1e-3)
    assert test_2.p.z == pytest.approx(2, rel=1e-3)
    assert test_2.v.x == pytest.approx(2, rel=1e-3)
    assert test_2.v.y == pytest.approx(0, rel=1e-3)
    assert test_2.v.z == pytest.approx(0, rel=1e-3)


def test_closest_point():
    """Test the Ray3D closest_point method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)

    near_pt = Point3D(3, 3, 0)
    assert ray.closest_point(near_pt) == Point3D(2, 3, 2)
    near_pt = Point3D(2, 0, 0)
    assert ray.closest_point(near_pt) == Point3D(2, 2, 2)
    near_pt = Point3D(1, 5, 0)
    assert ray.closest_point(near_pt) == Point3D(2, 5, 2)


def test_distance_to_point():
    """Test the Ray3D distance_to_point method."""
    pt = Point3D(2, 2, 2)
    vec = Vector3D(0, 2, 0)
    ray = Ray3D(pt, vec)

    near_pt = Point3D(3, 3, 2)
    assert ray.distance_to_point(near_pt) == 1
    near_pt = Point3D(2, 0, 2)
    assert ray.distance_to_point(near_pt) == 2
    near_pt = Point3D(1, 5, 2)
    assert ray.distance_to_point(near_pt) == 1


def test_to_from_array():
    """Test to/from array method."""
    test_ray = Ray3D(Point3D(2, 0, 2), Vector3D(2, 2, 2))
    ray_array = ((2, 0, 2), (2, 2, 2))

    assert test_ray == Ray3D.from_array(ray_array)

    ray_array = ((2, 0, 2), (2, 2, 2))
    test_ray = Ray3D(Point3D(2, 0, 2), Vector3D(2, 2, 2))

    assert test_ray.to_array() == ray_array

    test_ray_2 = Ray3D.from_array(test_ray.to_array())
    assert test_ray == test_ray_2



# coding=utf-8
"""Utility functions for computing intersections between geometry in 3D space."""
from __future__ import division

from .geometry3d.pointvector import Point3D

import math


def intersect_line3d_plane(line_ray, plane):
    """Get the intersection between a Ray3D/LineSegment3D and a Plane.

    Args:
        line_ray: A LineSegment3D or Ray3D object.
        plane: A Plane object to intersect.

    Returns:
        Point2D of intersection if it exists. None if no intersection exists.
    """
    d = plane.n.dot(line_ray.v)
    if not d:  # parallel
        return None
    u = (plane.k - plane.n.dot(line_ray.p)) / d
    if not line_ray._u_in(u):  # line or ray does not have its domain in the plane
        return None
    return Point3D(line_ray.p.x + u * line_ray.v.x,
                   line_ray.p.y + u * line_ray.v.y,
                   line_ray.p.z + u * line_ray.v.z)


def intersect_plane_plane(plane_a, plane_b):
    """Get the intersection between two Plane objects.

    Args:
        plane_a: A Plane object.
        plane_b: Another Plane object to intersect.

    Returns:
        Two objects that define the intersection between two planes

        1) A Point3D that lies along the intersection of the two planes.
        2) A Vector3D that describes the direction of the intersection.

        Will be None if no intersection exists (planes are parallel).
    """
    n1_m = plane_a.n.magnitude_squared
    n2_m = plane_b.n.magnitude_squared
    n1d2 = plane_a.n.dot(plane_b.n)
    det = n1_m * n2_m - n1d2 ** 2
    if det == 0:  # parallel
        return None
    c1 = (plane_a.k * n2_m - plane_b.k * n1d2) / det
    c2 = (plane_b.k * n1_m - plane_a.k * n1d2) / det
    return Point3D(c1 * plane_a.n.x + c2 * plane_b.n.x,
                   c1 * plane_a.n.y + c2 * plane_b.n.y,
                   c1 * plane_a.n.z + c2 * plane_b.n.z), plane_a.n.cross(plane_b.n)


def closest_point3d_on_line3d(point, line_ray):
    """Get the closest Point3D on a LineSegment3D or Ray3D to the input point.

    Args:
        point: A Point3D object.
        line_ray: A LineSegment3D or Ray3D object along which the closest point
            will be determined.

    Returns:
        Point3D for the closest point on line_ray to point.
    """
    d = line_ray.v.magnitude_squared
    assert d != 0, 'Length of LineSegment3D must be greater than 0.'
    u = ((point.x - line_ray.p.x) * line_ray.v.x +
         (point.y - line_ray.p.y) * line_ray.v.y +
         (point.z - line_ray.p.z) * line_ray.v.z) / d
    if not line_ray._u_in(u):
        u = max(min(u, 1.0), 0.0)
    return Point3D(line_ray.p.x + u * line_ray.v.x,
                   line_ray.p.y + u * line_ray.v.y,
                   line_ray.p.z + u * line_ray.v.z)


def closest_point3d_on_line3d_infinite(point, line_ray):
    """Get the closest Point3D on an infinite extension of a LineSegment3D or Ray3D.

    Args:
        point: A Point3D object.
        line_ray: A LineSegment3D or Ray3D object along which the closest point
            will be determined.

    Returns:
        Point3D for the closest point on the line_ray to the point.
    """
    d = line_ray.v.magnitude_squared
    assert d != 0, 'Length of LineSegment3D must be greater than 0.'
    u = ((point.x - line_ray.p.x) * line_ray.v.x +
         (point.y - line_ray.p.y) * line_ray.v.y +
         (point.z - line_ray.p.z) * line_ray.v.z) / d
    return Point3D(line_ray.p.x + u * line_ray.v.x,
                   line_ray.p.y + u * line_ray.v.y,
                   line_ray.p.z + u * line_ray.v.z)


def closest_point3d_on_plane(point, plane):
    """Get the closest Point3D on a Plane to the input point.

    Args:
        point: A Point3D object.
        plane: A Plane object in which the closest point will be determined.

    Returns:
        Point3D for the closest point on the plane to point.
    """
    n = plane.n
    d = point.dot(plane.n) - plane.k
    return Point3D(point.x - n.x * d, point.y - n.y * d, point.z - n.z * d)


def closest_point3d_between_line3d_plane(line_ray, plane):
    """Get the two closest Point3D between a LineSegment3D/Ray3D and a Plane.

    Args:
        line_ray: A LineSegment3D or Ray3D object along which the closest point
            will be determined.
        plane: A Plane object on which a closest point will be determined.

    Returns:
        Two Point3D objects representing

        1) The point on the line_ray that is closest to the plane.
        2) The point on the plane that is closest to the line_ray.

        Will be None if there is an intersection between line_ray and the plane
    """
    d = plane.n.dot(line_ray.v)
    if not d:  # parallel, choose an endpoint
        return line_ray.p, closest_point3d_on_plane(line_ray.p, plane)
    u = (plane.k - plane.n.dot(line_ray.p)) / d
    if not line_ray._u_in(u):  # intersects out of range of L, choose nearest endpoint
        u = max(min(u, 1.0), 0.0)
        close_pt = Point3D(line_ray.p.x + u * line_ray.v.x,
                           line_ray.p.y + u * line_ray.v.y,
                           line_ray.p.z + u * line_ray.v.z)
        return close_pt, closest_point3d_on_plane(close_pt, plane)
    return None  # intersection


def intersect_line3d_sphere(line_ray, sphere):
    """Get the intersection between this Sphere object and a Ray2D/LineSegment2D.

    Args:
        line_ray: A LineSegment3D or Ray3D for intersection.
        sphere: A Sphere to intersect.

    Returns:
        Two Point3D objects if a full intersection exists.
        A Point3D if a point of tangency exists.
        Will be None if no intersection exists.

    """
    L = line_ray
    S = sphere
    a = L.v.magnitude_squared
    b = 2 * (L.v.x * (L.p.x - S.center.x) +
             L.v.y * (L.p.y - S.center.y) +
             L.v.z * (L.p.z - S.center.z))
    c = S.center.magnitude_squared + \
        L.p.magnitude_squared - \
        2 * S.center.dot(L.p) - \
        S.radius ** 2
    det = b ** 2 - 4 * a * c
    if det < 0:
        return None
    sq = math.sqrt(det)
    u1 = (-b + sq) / (2 * a)
    u2 = (-b - sq) / (2 * a)
    if not L._u_in(u1):
        u1 = max(min(u1, 1.0), 0.0)
    if not L._u_in(u2):
        u2 = max(min(u2, 1.0), 0.0)
    p1 = Point3D(L.p.x + u1 * L.v.x, L.p.y + u1 * L.v.y, L.p.z + u1 * L.v.z)
    if u1 == u2:
        return p1
    else:
        p2 = Point3D(L.p.x + u2 * L.v.x, L.p.y + u2 * L.v.y, L.p.z + u2 * L.v.z)
        return p1, p2


def intersect_plane_sphere(plane, sphere):
    """Get the intersection of a plane with this Sphere object

    Args:
        plane: A Plane object.
        sphere: A Sphere to intersect.

    Returns:
        If a full intersection exists

        1) A Point3D that represents the center of the intersection circle.
        2) A Vector3D that represents the normal of the intersection circle.
        3) A number that represents the radius of the intersection circle.

        A Point3D Object if a point of tangency exists.
        None if no intersection exists.
    """
    r = sphere.radius
    pt_c = sphere.center
    pt_o = plane.o
    v_n = plane.n.normalize()

    # Resulting circle radius. Radius² = r² - [(c-p).n]²
    d = (pt_o - pt_c).dot(v_n)
    if abs(r) < abs(d):  # No intersection if (r ** 2 - d ** 2) negative
        return None
    cut_r = math.sqrt(r ** 2 - d ** 2)

    # Intersection circle center point. Center_point = p - [(c-p).n]n
    cut_center = pt_c + (d * v_n)

    return (cut_center, v_n, cut_r) if cut_r != 0 else cut_center



# coding=utf-8
import pytest

from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.ray import Ray2D

import math


def test_ray2d_init():
    """Test the initalization of Ray2D objects and basic properties."""
    pt = Point2D(2, 0)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)
    str(ray)  # test the string representation of the ray

    assert ray.p == Point2D(2, 0)
    assert ray.v == Vector2D(0, 2)

    flip_ray = ray.reverse()
    assert flip_ray.p == Point2D(2, 0)
    assert flip_ray.v == Vector2D(0, -2)


def test_ray2_to_from_dict():
    """Test the to/from dict of Ray2D objects."""
    pt = Point2D(2, 0)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)
    ray_dict = ray.to_dict()
    new_ray = Ray2D.from_dict(ray_dict)
    assert isinstance(new_ray, Ray2D)
    assert new_ray.to_dict() == ray_dict


def test_ray2d_immutability():
    """Test the immutability of Ray2D objects."""
    pt = Point2D(2, 0)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)

    assert isinstance(ray, Ray2D)
    with pytest.raises(AttributeError):
        ray.p.x = 3
    with pytest.raises(AttributeError):
        ray.v.x = 3
    with pytest.raises(AttributeError):
        ray.p = Point2D(0, 0)
    with pytest.raises(AttributeError):
        ray.v = Vector2D(2, 2)

    ray_copy = ray.duplicate()
    assert ray.p == ray_copy.p
    assert ray.v == ray_copy.v


def test_equality():
    """Test the equality of Ray2D objects."""
    pt = Point2D(2, 0)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)
    ray_dup = ray.duplicate()
    ray_alt = Ray2D(Point2D(2, 0.1), vec)

    assert ray is ray
    assert ray is not ray_dup
    assert ray == ray_dup
    assert hash(ray) == hash(ray_dup)
    assert ray != ray_alt
    assert hash(ray) != hash(ray_alt)


def test_move():
    """Test the Ray2D move method."""
    pt = Point2D(2, 0)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)

    vec_1 = Vector2D(2, 2)
    new_ray = ray.move(vec_1)
    assert new_ray.p == Point2D(4, 2)
    assert new_ray.v == vec


def test_scale():
    """Test the Ray2D scale method."""
    pt = Point2D(2, 2)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)

    origin_1 = Point2D(0, 2)
    origin_2 = Point2D(1, 1)
    new_ray = ray.scale(2, origin_1)
    assert new_ray.p == Point2D(4, 2)
    assert new_ray.v == Point2D(0, 4)
    assert new_ray.v.magnitude == 4

    new_ray = ray.scale(2, origin_2)
    assert new_ray.p == Point2D(3, 3)
    assert new_ray.v == Point2D(0, 4)


def test_scale_world_origin():
    """Test the Ray2D scale method with None origin."""
    pt = Point2D(2, 2)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)

    new_ray = ray.scale(2)
    assert new_ray.p == Point2D(4, 4)
    assert new_ray.v == Point2D(0, 4)
    assert new_ray.v.magnitude == 4


def test_rotate():
    """Test the Ray2D rotate method."""
    pt = Point2D(2, 2)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)
    origin_1 = Point2D(0, 2)

    test_1 = ray.rotate(math.pi, origin_1)
    assert test_1.p.x == pytest.approx(-2, rel=1e-3)
    assert test_1.p.y == pytest.approx(2, rel=1e-3)
    assert test_1.v.x == pytest.approx(0, rel=1e-3)
    assert test_1.v.y == pytest.approx(-2, rel=1e-3)

    test_2 = ray.rotate(math.pi/2, origin_1)
    assert test_2.p.x == pytest.approx(0, rel=1e-3)
    assert test_2.p.y == pytest.approx(4, rel=1e-3)
    assert test_2.v.x == pytest.approx(-2, rel=1e-3)
    assert test_2.v.y == pytest.approx(0, rel=1e-3)


def test_reflect():
    """Test the Ray2D reflect method."""
    pt = Point2D(2, 2)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)

    origin_1 = Point2D(0, 1)
    origin_2 = Point2D(1, 1)
    normal_1 = Vector2D(0, 1)
    normal_2 = Vector2D(-1, 1).normalize()

    assert ray.reflect(normal_1, origin_1).p == Point2D(2, 0)
    assert ray.reflect(normal_1, origin_1).v == Vector2D(0, -2)
    assert ray.reflect(normal_1, origin_2).p == Point2D(2, 0)
    assert ray.reflect(normal_1, origin_2).v == Vector2D(0, -2)

    test_1 = ray.reflect(normal_2, origin_2)
    assert test_1.p == Point2D(2, 2)
    assert test_1.v.x == pytest.approx(2, rel=1e-3)
    assert test_1.v.y == pytest.approx(0, rel=1e-3)

    test_2 = ray.reflect(normal_2, origin_1)
    assert test_2.p.x == pytest.approx(1, rel=1e-3)
    assert test_2.p.y == pytest.approx(3, rel=1e-3)
    assert test_1.v.x == pytest.approx(2, rel=1e-3)
    assert test_1.v.y == pytest.approx(0, rel=1e-3)


def test_closest_point():
    """Test the Ray2D closest_point method."""
    pt = Point2D(2, 2)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)

    near_pt = Point2D(3, 3)
    assert ray.closest_point(near_pt) == Point2D(2, 3)
    near_pt = Point2D(2, 0)
    assert ray.closest_point(near_pt) == Point2D(2, 2)
    near_pt = Point2D(1, 5)
    assert ray.closest_point(near_pt) == Point2D(2, 5)


def test_distance_to_point():
    """Test the Ray2D distance_to_point method."""
    pt = Point2D(2, 2)
    vec = Vector2D(0, 2)
    ray = Ray2D(pt, vec)

    near_pt = Point2D(3, 3)
    assert ray.distance_to_point(near_pt) == 1
    near_pt = Point2D(2, 0)
    assert ray.distance_to_point(near_pt) == 2
    near_pt = Point2D(1, 5)
    assert ray.distance_to_point(near_pt) == 1


def test_intersect_line_ray():
    """Test the Ray2D distance_to_point method."""
    pt_1 = Point2D(2, 2)
    vec_1 = Vector2D(0, 2)
    ray_1 = Ray2D(pt_1, vec_1)

    pt_2 = Point2D(0, 3)
    vec_2 = Vector2D(4, 0)
    ray_2 = Ray2D(pt_2, vec_2)

    pt_3 = Point2D(0, 0)
    vec_3 = Vector2D(1, 1)
    ray_3 = Ray2D(pt_3, vec_3)

    assert ray_1.intersect_line_ray(ray_2) == Point2D(2, 3)
    assert ray_1.intersect_line_ray(ray_3) == Point2D(2, 2)


def test_to_from_array():
    """Test to/from array method."""
    test_ray = Ray2D(Point2D(2, 0), Vector2D(2, 2))
    ray_array = ((2, 0), (2, 2))

    assert test_ray == Ray2D.from_array(ray_array)

    ray_array = ((2, 0), (2, 2))
    test_ray = Ray2D(Point2D(2, 0), Vector2D(2, 2))

    assert test_ray.to_array() == ray_array

    test_ray_2 = Ray2D.from_array(test_ray.to_array())
    assert test_ray == test_ray_2



# coding=utf-8
"""Base class for all 1D geometries in 2D space (Ray2D and LineSegment2D)."""
from __future__ import division

from .pointvector import Vector2D, Point2D
from ..intersection2d import intersect_line2d, closest_point2d_on_line2d


class Base1DIn2D(object):
    """Base class for all 1D geometries in 2D space (Ray2D and LineSegment2D).

    Args:
        p: A Point2D representing the base.
        v: A Vector2D representing the direction.

    Properties:
        * p
        * v
    """
    __slots__ = ('_p', '_v')

    def __init__(self, p, v):
        """Initialize Base1DIn2D."""
        assert isinstance(p, Point2D), "Expected Point2D. Got {}.".format(type(p))
        assert isinstance(v, Vector2D), "Expected Vector2D. Got {}.".format(type(v))
        self._p = p
        self._v = v

    @classmethod
    def from_dict(cls, data):
        """Create a LineSegment2D/Ray2D from a dictionary.

        Args:
            data: A python dictionary in the following format

        .. code-block:: python

            {
            "p": (10, 0),
            "v": (10, 10)
            }
        """
        return cls(Point2D.from_array(data['p']),
                   Vector2D.from_array(data['v']))

    @property
    def p(self):
        """Base point."""
        return self._p

    @property
    def v(self):
        """Direction vector."""
        return self._v

    def closest_point(self, point):
        """Get the closest Point2D on this object to another Point2D.

        Args:
            point: A Point2D object to which the closest point on this object
                will be computed.

        Returns:
            Point2D for the closest point on this line to the input point.
        """
        return closest_point2d_on_line2d(point, self)

    def distance_to_point(self, point):
        """Get the minimum distance between this object and the input point.

        Args:
            point: A Point2D object to which the minimum distance will be computed.

        Returns:
            The distance to the input point.
        """
        close_pt = self.closest_point(point)
        return point.distance_to_point(close_pt)

    def intersect_line_ray(self, line_ray):
        """Get the intersection between this object and another Ray2 or LineSegment2D.

        Args:
            line_ray: Another LineSegment2D or Ray2D or to intersect.

        Returns:
            Point2D of intersection if it exists. None if no intersection exists.
        """
        return intersect_line2d(self, line_ray)

    def duplicate(self):
        """Get a copy of this object."""
        return self.__copy__()

    def to_dict(self):
        """Get LineSegment2D/Ray2D as a dictionary."""
        return {'p': self.p.to_array(),
                'v': self.v.to_array()}

    def __copy__(self):
        return self.__class__(self.p, self.v)

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (hash(self.p), hash(self.v))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Base1DIn2D) and self.__key() == other.__key()
    
    def __ne__(self, other):
        return not self.__eq__(other)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """Base1Din2D representation."""
        return 'Base 1D Object (2D Space)'
